package com.genpt.nsight.v4.service.impl;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.genpt.nsight.NSightCache;
import com.genpt.nsight.v1.model.SiteIdentifier;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v1.model.serializer.DeliveryCutOffSerializer;
import com.genpt.nsight.v4.model.RawDelivery;
import com.genpt.nsight.v4.service.delivery.DeliveryOption;
import com.genpt.nsight.v4.service.delivery.DeliveryResponseV1;
import com.google.api.core.ApiFuture;
import com.google.cloud.pubsub.v1.Publisher;
import com.google.common.base.Stopwatch;
import com.google.protobuf.ByteString;
import com.google.pubsub.v1.PubsubMessage;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.time.DurationFormatUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

@Service
@Slf4j
@ConditionalOnExpression("${delivery.enablePublishFreeDelivery:false}")
public class AsyncDeliveryPubsubPublisher {
    @Autowired
    private Publisher publisher;
    private ObjectMapper objectMapper;

    @Autowired
    NSightCache nSightCache;

    @PostConstruct
    public void init(){
        objectMapper =  new ObjectMapper();
    }

    private void publishMessage(DeliveryResponseV1 deliveryResponse, String destination, Optional<String> sources) throws JsonProcessingException, ExecutionException, InterruptedException {
        String response = objectMapper.writeValueAsString(deliveryResponse);
        ByteString data = ByteString.copyFrom(response.getBytes());
        PubsubMessage message = null;
        if(sources.isPresent())
            message = PubsubMessage.newBuilder().setData(data).putAttributes("X-Destination", destination).putAttributes("X-Sources", String.valueOf(sources)).build();
        else
            message = PubsubMessage.newBuilder().setData(data).putAttributes("X-Destination", destination).build();
        ApiFuture<String>  future = publisher.publish(message);
        String ackId = future.get();
    }

    @Async("asyncDeliveryExecutor")
    public CompletableFuture<String> triggerPublish(Map<String, List<RawDelivery>> deliveryResponseMap, String destination, Optional<String> sources) throws Exception {
        // now publish the message to pubsub
        try {
            Stopwatch stopwatch = Stopwatch.createStarted();
            DeliveryResponseV1 deliveryResponseV1 = new DeliveryResponseV1();
            deliveryResponseV1.setDeliveryOptions(new ArrayList<>());

            //need to iterate over the map and the deliveries and create the S&D response format
            for (Map.Entry<String,List<RawDelivery>> entry : deliveryResponseMap.entrySet())  {
                // the key is the source site id and the value is a list of deliveries
                //need to get the location ID from the key/siteid
                SiteDTO siteDTOSource = nSightCache.getSiteIdToSiteMap().get(entry.getKey());
                if (siteDTOSource != null && siteDTOSource.getSiteIdentifierList() != null) {
                    Optional<SiteIdentifier> siteIdentifierSource = siteDTOSource.getSiteIdentifierList().stream().filter(siteIdentifier -> siteIdentifier.getName().contains("9 Digit Store Number") || siteIdentifier.getName().contains("DC Abbreviation Code")).findAny();
                    if (siteIdentifierSource.isPresent()) {
                        String source = siteIdentifierSource.get().getValue();
                        entry.getValue().forEach(rawDelivery -> {
                            rawDelivery.getDeliveryCutOff().forEach(deliveryCutOff -> {
                                //for each delivery create a S&D delivery Option
                                DeliveryOption deliveryOption = new DeliveryOption();
                                deliveryOption.setDestination(rawDelivery.getDeliveryDestination().getValue());
                                deliveryOption.setSource(source);
                                deliveryOption.setDay(DeliveryCutOffSerializer.DAY_OF_WEEK[ deliveryCutOff.getDayOfTheWeek()]);
                                if(deliveryCutOff.getCutOffTime()!=-1) {
                                    //deliveryOption.setCutoff(DurationFormatUtils.formatDuration(deliveryCutOff.getCutOffTime(), "HH:MM:SS", true));
                                    Duration duration = Duration.ofMillis(deliveryCutOff.getCutOffTime());
                                    long seconds = duration.getSeconds();
                                    long HH = seconds / 3600;
                                    long MM = (seconds % 3600) / 60;
                                    long SS = seconds % 60;
                                    deliveryOption.setCutoff(String.format("%02d:%02d:%02d", HH, MM, SS));
                                }
                                if(deliveryCutOff.getEarliesTime()!=-1) {
                                    Duration duration = Duration.ofMillis(deliveryCutOff.getEarliesTime());
                                    long seconds = duration.getSeconds();
                                    long HH = seconds / 3600;
                                    long MM = (seconds % 3600) / 60;
                                    long SS = seconds % 60;
                                    deliveryOption.setEarliestTime(String.format("%02d:%02d:%02d", HH, MM, SS));
                                }
                                if(deliveryCutOff.getLatestTime()!=-1) {
                                    Duration duration = Duration.ofMillis(deliveryCutOff.getLatestTime());
                                    long seconds = duration.getSeconds();
                                    long HH = seconds / 3600;
                                    long MM = (seconds % 3600) / 60;
                                    long SS = seconds % 60;
                                    deliveryOption.setLatestTime(String.format("%02d:%02d:%02d", HH, MM, SS));
                                }
                                //deliveryOption.setDeliveryName("");//dont have this
                                deliveryOption.setCategoryType(rawDelivery.getDeliveryCategory());
                                deliveryOption.setDurationMaxValue(deliveryCutOff.getDeliveryEstimate().getDurationMaxValue());
                                deliveryOption.setDurationMinValue(deliveryCutOff.getDeliveryEstimate().getDurationMinValue());
                                //deliveryOption.setDataAsOfTs();
                                deliveryOption.setDurationType(deliveryCutOff.getDeliveryEstimate().getDurationDesignator());
                                deliveryOption.setServiceType(rawDelivery.getDeliveryType());
                                //deliveryOption.setSourceDataAsOfTs("");
                                deliveryResponseV1.getDeliveryOptions().add(deliveryOption);

                            });//each cutoff
                        });//each delivery
                    }
                }
            }
            //now publish the message
            try {
                publishMessage(deliveryResponseV1, destination,sources);
            } catch (Exception e) {
                log.error("Exception Publishing Delivery message to Pubsub, Exception {}, Message {}",e, e.getMessage());
            }
            log.info("Published Delivery Response message to Pubsub in {}", stopwatch.stop());
        } catch (Exception e) {
            log.error("Exception Transforming Delivery message to Pubsub, Exception {}, Message {}",e, e.getMessage());
        }
        return CompletableFuture.completedFuture("Success");
    }
}
