package com.genpt.nsight.v4.service.impl;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.genpt.nsight.v4.model.pubsub.Action;
import com.genpt.nsight.v4.model.pubsub.InventoryEvent;
import com.genpt.nsight.v4.model.pubsub.LocationType;
import com.google.api.core.ApiFuture;
import com.google.cloud.pubsub.v1.Publisher;
import com.google.protobuf.ByteString;
import com.google.pubsub.v1.PubsubMessage;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

@Service
@Slf4j
@ConditionalOnExpression("${nsight.VerificationControllerV2.enabled:false}")
public class AsyncPubsubPublisher {
    @Autowired
    private Publisher publisher;
    private ObjectMapper objectMapper;

    @PostConstruct
    public void init(){
        objectMapper =  new ObjectMapper();
    }

    public LocationType getLocationType(String locationTypeString) {
        LocationType locationType= LocationType.STORE;
        if(locationTypeString.equalsIgnoreCase(LocationType.DC.getLocationType()))
            locationType = LocationType.DC;
        return locationType;
    }

    private void publishMessage(InventoryEvent inventoryEvent, String correlationId) throws JsonProcessingException, ExecutionException, InterruptedException {
        String request = objectMapper.writeValueAsString(inventoryEvent);
        ByteString data = ByteString.copyFrom(request.getBytes());
        PubsubMessage message = PubsubMessage.newBuilder().setData(data).putAttributes("CorrelationId", correlationId).build();
        ApiFuture<String>  future = publisher.publish(message);
        String ackId = future.get();

    }
    @Async("asyncExecutor")
    public CompletableFuture<String> triggerPublish(com.genpt.nsight.v2.model.ProductAvailabilityResponse finalProductAvailabilityResponse) throws Exception {
        long timestamp = finalProductAvailabilityResponse.getHeaderResponse().getTimestamp();
        String correlationId = finalProductAvailabilityResponse.getCorrelationId();
        finalProductAvailabilityResponse.getDetailResponse().getParts().forEach(part -> {
            InventoryEvent inventoryEvent = new InventoryEvent();
            // need to set the action type and location type
            inventoryEvent.setProductCode(part.getLineAbbrev().trim());
            inventoryEvent.setProductNumber(part.getPartNumber().trim());
            inventoryEvent.setCountryCode(finalProductAvailabilityResponse.getHeaderResponse().getCountryCode());
            if (part.getSites()!=null && !part.getSites().isEmpty()){
                //set the action accordingly
                com.genpt.nsight.v4.model.pubsub.Action action = Action.UPDATE;
                //for verification should be just one site
                part.getSites().forEach(site -> {// should be just one site
                    inventoryEvent.setAction(action);
                    inventoryEvent.setLocationType(getLocationType(site.getSiteType()));
                    //we should only be getting this for Store and DC
                    if(inventoryEvent.getLocationType().equals(LocationType.STORE))
                        inventoryEvent.setLocation(site.getSiteIdentifiers().stream().filter(siteIdentifier -> siteIdentifier.getName().compareToIgnoreCase("9 Digit Store Number")==0).findFirst().get().getValue());
                    else
                        inventoryEvent.setLocation(site.getSiteIdentifiers().stream().filter(siteIdentifier -> siteIdentifier.getName().compareToIgnoreCase("DC Abbreviation Code")==0).findFirst().get().getValue());
                    inventoryEvent.setAvailQty(site.getQuantity().getAvailable());
                    inventoryEvent.setInTransitTs(site.getQuantity().getInTransitDate());
                    inventoryEvent.setDataAsOfTs(site.getSourceSystemDataAsOfDate()>0 ? site.getSourceSystemDataAsOfDate() : timestamp);
                    inventoryEvent.setInTransitQty(site.getQuantity().getInTransit());
                    inventoryEvent.setIncrSellQty(site.getQuantity().getIncrementalSellQuantity());
                    inventoryEvent.setMaxStockQty(site.getQuantity().getMaxStockedQuantity());
                    inventoryEvent.setMinStockQty(site.getQuantity().getMinStockedQuantity());
                    inventoryEvent.setMinSellQty(site.getQuantity().getMinimumSellQuantity());
                    inventoryEvent.setOnHandQty(site.getQuantity().getOnHand());
                    inventoryEvent.setOnOrderQty(site.getQuantity().getOnOrder());
                    inventoryEvent.setOnOrderTs(site.getQuantity().getOnOrderDate());
                    inventoryEvent.setOrderSinceLastUpdate(site.getOrderSinceLastUpdate());
                    inventoryEvent.setProductDescr(site.getPartAttributes().getPartDescription() == null ? "" : site.getPartAttributes().getPartDescription());
                    inventoryEvent.setUom(site.getPartAttributes().getUnitOfMeasure());
                });
                // now publish the message to pubsub
                try {
                    publishMessage(inventoryEvent, correlationId);
                    log.info("Published Update message to Pubsub for correlation ID  {}", correlationId);
                } catch (Exception e) {
                    log.error("Exception Publishing Update  message to Pubsub for correlation ID  {} , Exception {}, Message {}",correlationId,e, e.getMessage());
                }
            }else{// part not found
                //we dont need to publish anything for part not found
            }
        });
        return CompletableFuture.completedFuture("Success");
    }
}
