package com.genpt.nsight.v4.service.impl;

import com.datastax.driver.core.Session;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.genpt.nsight.NSightCache;
import com.genpt.nsight.v1.model.*;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v2.SplunkLogger;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.model.ProductAvailabilityRequest;
import com.genpt.nsight.v4.model.StoreHours;
import com.genpt.nsight.v4.model.inventory.availability.InventoryResponseMultiProduct;
import com.genpt.nsight.v4.model.inventory.availability.InventoryResponseV1;
import com.genpt.nsight.v4.service.*;
import com.genpt.nsight.v4.service.delivery.DeliveryService;
import com.genpt.nsight.v4.service.inventory.AsyncInventoryComparator;
import com.google.common.base.Stopwatch;
import com.google.common.collect.ImmutableMap;
import com.google.common.util.concurrent.ListenableFuture;
import lombok.Getter;
import lombok.Setter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Configuration;
import org.springframework.jms.connection.CachingConnectionFactory;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.stream.Collectors;

@Service("ProductAvailabilityDetailServiceV4")
@Getter
@Setter
@Configuration
public class ProductAvailabilityDetailServiceV4Impl implements ProductAvailabilityDetailServiceV4 {
    @Autowired
    CachingConnectionFactory jmsConnectionFactory;
    private static final Logger LOGGER = LoggerFactory.getLogger(ProductAvailabilityDetailServiceV4Impl.class);

    @Autowired
    private Session session;
    @Autowired
    private NSightCache nSightCache;
    @Autowired
    SiteRelationshipServiceV4 siteRelationshipServiceV3;
    @Autowired
    private InventoryServiceV4 inventoryServiceV2;
    @Autowired
    private Site2SiteDeliveryEstimateServiceV4 deliveryEstimateService;
    @Autowired
    private Site2ZipDeliveryEstimateServiceV4 deliveryEstimateServiceByZip;

    @Autowired
    private LeadTimesServiceV4 leadTimesServiceV3;

    @Autowired
    ObjectMapper objectMapper;

    private Tracking tracking = new Tracking();

    private List<Log> noPartsFound = Collections.singletonList(new Log("2001","No Records found in Inventory Tables"));
    private List<Log> partNotFound = Collections.singletonList(new Log("2001","Part Not Found"));
    private List<Log> priceUnavailableDC = Collections.singletonList(new Log("2001","Pricing Data Unavailable for DC"));
    private List<Log> priceUnavailableSupplier = Collections.singletonList(new Log("2001","Pricing Data Unavailable for Supplier"));
    private List<Log> priceUnavailableStore = Collections.singletonList(new Log("2001","Unavailable to Retrieve Pricing Data for Store"));
    @Value("${inventory.allCallsToGcp}")
    private boolean allCallsToGcp;

    @Value("${inventory.allCallsToNsight}")
    private boolean allCallsToNsight;
    @Value("${inventory.invDataSvcPilotSellingChannels:123}")
    private String invDataSvcPilotSellingChannels;

    @Value("${inventory.pilotSellingChannelsCallsToGcp}")
    private boolean pilotSellingChannelsCallsToGcp;
    @Value("${inventory.pilotSellingChannelsResponseFromGcp}")
    private boolean pilotSellingChannelsResponseFromGcp;
    @Value("${inventory.enableCompare}")
    private boolean enableCompare;

    @Value("${inventory.useMultiProductEndpoint}")
    private boolean useMultiProductEndpoint;
    @Value("${inventory.partCountCallToGCP}")
    private int partCountCallToGCP;

    @Autowired
    AsyncInventoryComparator asyncInventoryComparator;

    private void handleAvailabilityCalls(AvailabilityServiceContextV4 availabilityServiceContextV3, SplunkLogger splunkLogger) {
        long queryStartTime;
        queryStartTime = System.currentTimeMillis();
        //probably send the BT requests first before the cassandra requests
        if (allCallsToGcp) {
            Map<String, List<com.genpt.nsight.v4.model.Part>> inventoryMap = null;
            long invResponseProcessingStartTime = System.currentTimeMillis();
            if(useMultiProductEndpoint){
                ListenableFuture<InventoryResponseMultiProduct> gcpMultiProductFutures = inventoryServiceV2.createAvailabilityCallsMultiPart(availabilityServiceContextV3);
                availabilityServiceContextV3.setGcpMultiProductInventoryFutures(gcpMultiProductFutures);
                LOGGER.debug(splunkLogger.toString() + " Submitted Inventory Domain MultiProduct Availability Inventory queries in {} ms", (System.currentTimeMillis() - queryStartTime));
                //Set if we should or not  use the hqline abbrev logic
                setSkipHQLineAbbrev(availabilityServiceContextV3);
                inventoryMap = inventoryServiceV2.processMultiProductAsyncResponses(
                        availabilityServiceContextV3);
            }else {
                ListenableFuture<List<InventoryResponseV1>> gcpResponseFutures =
                        inventoryServiceV2.createAvailabilityCallsAsync(availabilityServiceContextV3);
                availabilityServiceContextV3.setGcpInventoryFutures(gcpResponseFutures);
                LOGGER.debug(splunkLogger.toString() + " Submitted Inventory Domain Availability Inventory queries in {} ms", (System.currentTimeMillis() - queryStartTime));
                //Set if we should or not  use the hqline abbrev logic
                setSkipHQLineAbbrev(availabilityServiceContextV3);
                inventoryMap = inventoryServiceV2.processAvailabilityAsyncResponses(
                        availabilityServiceContextV3);
            }
            availabilityServiceContextV3.setGcpInventory(inventoryMap);
            availabilityServiceContextV3.setInventory(inventoryMap);
            LOGGER.info(splunkLogger.toString() + " All Calls to GCP, Processed GCP Inventory Query Response in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        } else if (allCallsToNsight) {
            //Set Inventory Futures for Cassandra
            inventoryServiceV2.setInventoryFutures(availabilityServiceContextV3);
            LOGGER.debug(
                    splunkLogger.toString() + " Submitted Inventory queries in {} ms", (System.currentTimeMillis() - queryStartTime));
            long invResponseProcessingStartTime = System.currentTimeMillis();
            //Set if we should or not  use the hqline abbrev logic
            setSkipHQLineAbbrev(availabilityServiceContextV3);

            inventoryServiceV2.processInventoryFutures(availabilityServiceContextV3);
            LOGGER.info(splunkLogger.toString() + " All Calls to NSight, Processed NSight Inventory Query Response in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));

        } else {
            //mix and match the calls. for now for a given set of stores/clientId/single part
            //make the calls to GCP
            ListenableFuture<List<InventoryResponseV1>> gcpResponseFutures = null;
            ListenableFuture<InventoryResponseMultiProduct> gcpMultiProductFutures = null;
            Optional<SiteIdentifier> locationSiteId = availabilityServiceContextV3.getRequestingSite().getSiteIdentifierList()
                    .stream()
                    .filter(siteIdentifier -> siteIdentifier.getName().compareToIgnoreCase("9 Digit Store Number") == 0 || siteIdentifier.getName().compareToIgnoreCase("DC Abbreviation Code") == 0).findFirst();
            Optional<String> pilotLocationPrefix = Optional.empty();
            Optional<String> pilotLocation = Optional.empty();
            String pilotLocationPrefixAll = availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode()+"-ALL";
            if(locationSiteId.isPresent() && locationSiteId.get().getValue().length() > 2) {
                pilotLocationPrefix = Optional.of(availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode() + "-" + locationSiteId.get().getValue().substring(0, 3));
                pilotLocation = Optional.of(availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode() + "-" + locationSiteId.get().getValue());
            }
            boolean isGCPRequestNeeded =(pilotSellingChannelsCallsToGcp && invDataSvcPilotSellingChannels.contains(availabilityServiceContextV3.getApplicationID())
                    && availabilityServiceContextV3.getPartSet().size() <= partCountCallToGCP)
                    ||
                    (pilotLocationPrefix.isPresent()
                            && (nSightCache.getInvDataSvcPilotStores().toUpperCase().contains(pilotLocationPrefix.get().toUpperCase())
                            || nSightCache.getInvDataSvcPilotStores().toUpperCase().contains(pilotLocationPrefixAll.toUpperCase())
                            || nSightCache.getInvDataSvcPilotStores().toUpperCase().contains(pilotLocation.get().toUpperCase()))
                            && availabilityServiceContextV3.getPartSet().size() <= partCountCallToGCP);
            //make calls to nsight only if data from nsight is required
            if(isEnableCompare() || !isGCPRequestNeeded || !pilotSellingChannelsResponseFromGcp){
                inventoryServiceV2.setInventoryFutures(availabilityServiceContextV3);
                LOGGER.debug(
                        splunkLogger.toString() + " Submitted NSight Queries in {} ms", (System.currentTimeMillis() - queryStartTime));
            }
            if (isGCPRequestNeeded) {
                //calls goto GCP either to single part endpoint or multipart
                if(useMultiProductEndpoint){
                    gcpMultiProductFutures = inventoryServiceV2.createAvailabilityCallsMultiPart(availabilityServiceContextV3);
                    availabilityServiceContextV3.setGcpMultiProductInventoryFutures(gcpMultiProductFutures);
                    LOGGER.debug(splunkLogger.toString() + " Submitted Async calls to GCP Multi Product endpoint in {} ms", (System.currentTimeMillis() - queryStartTime));
                }else {
                    gcpResponseFutures = inventoryServiceV2.createAvailabilityCallsAsync(availabilityServiceContextV3);
                    availabilityServiceContextV3.setGcpInventoryFutures(gcpResponseFutures);
                    LOGGER.debug(splunkLogger.toString() + " Submitted Async calls to GCP Single Product endpoint in {} ms", (System.currentTimeMillis() - queryStartTime));
                }
            }
            long invResponseProcessingStartTime = System.currentTimeMillis();
            //Set if we should or not  use the hqline abbrev logic
            setSkipHQLineAbbrev(availabilityServiceContextV3);
            Map<String, List<com.genpt.nsight.v4.model.Part>> nsightInventoryMap = null;
            if(isEnableCompare() || !isGCPRequestNeeded || !pilotSellingChannelsResponseFromGcp){
                nsightInventoryMap = inventoryServiceV2.processInventoryFutures(availabilityServiceContextV3);
                LOGGER.info(splunkLogger.toString() + " Processed NSight Inventory Query Response in {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
            }
            //we will process the GCP responses asynchronously
            try {
                if(gcpResponseFutures!=null && !useMultiProductEndpoint) {
                    handleAndCompareSingleProductGCPResponses(availabilityServiceContextV3, splunkLogger,gcpResponseFutures,nsightInventoryMap);
                }
                if(gcpMultiProductFutures!=null && useMultiProductEndpoint) {
                    handleAndCompareMultiProductGCPResponses(availabilityServiceContextV3, splunkLogger,gcpMultiProductFutures,nsightInventoryMap);
                }
            } catch (Exception e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    private void handleAndCompareSingleProductGCPResponses(AvailabilityServiceContextV4 availabilityServiceContextV3,
                                                           com.genpt.nsight.v2.SplunkLogger splunkLogger,
                                                           ListenableFuture<List<InventoryResponseV1>> gcpResponseFutures,
                                                           Map<String, List<com.genpt.nsight.v4.model.Part>> nsightInventoryMap){
        long invResponseProcessingStartTime = System.currentTimeMillis();
        if(pilotSellingChannelsResponseFromGcp){// need the data from GCP in final response
            availabilityServiceContextV3.getSitesWithInventory().clear();

            Map<String, List<com.genpt.nsight.v4.model.Part>> gcpInventoryMap =
                    asyncInventoryComparator.getGCPInventory(gcpResponseFutures, availabilityServiceContextV3.getRequest());
            availabilityServiceContextV3.setGcpInventory(gcpInventoryMap);
            availabilityServiceContextV3.setInventory(gcpInventoryMap);
            List<List<com.genpt.nsight.v4.model.Part>> list = gcpInventoryMap.values().stream().map(parts -> parts.stream().collect(Collectors.toList())).collect(Collectors.toList());
            List<com.genpt.nsight.v4.model.Part> flat = list.stream()
                    .flatMap(List::stream)
                    .collect(Collectors.toList());
            availabilityServiceContextV3.setSiteWithPartsAvailable(flat.stream().map(part -> part.getSite().getSiteId()).collect(Collectors.toSet()));
            availabilityServiceContextV3.setSitesWithInventory(flat.stream().map(part -> part.getSite().getSiteId()).collect(Collectors.toSet()));

            LOGGER.info(splunkLogger.toString() + " Processed GCP Single Product Inventory Response for Configured Selling Channels in {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
            //at this point we can trigger an async request to do the compare
            // using nsightInventoryMap and gcpInventoryMap
            if(enableCompare){
                invResponseProcessingStartTime = System.currentTimeMillis();
                asyncInventoryComparator.compareResponsesAsync(ImmutableMap.copyOf(gcpInventoryMap),ImmutableMap.copyOf(nsightInventoryMap),new com.genpt.nsight.v1.SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest()));
                LOGGER.debug(splunkLogger + " Started Only GCP/Nsight Single Product Response Comparing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
            }else
                LOGGER.debug(splunkLogger + " Skipped GCP/Nsight Single Product Response Comparing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        }else{//GCP responses only for async compare with NSight
            //only async processing of futures and compare. no need to serve GCP in response
            invResponseProcessingStartTime = System.currentTimeMillis();
            asyncInventoryComparator.transfromGCPFutures(ImmutableMap.copyOf(availabilityServiceContextV3.getInventory()), gcpResponseFutures, availabilityServiceContextV3.getRequest(), enableCompare);
            LOGGER.debug(splunkLogger + " Started GCP/Nsight Single Product Response Comparing Processing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        }
    }

    private void handleAndCompareMultiProductGCPResponses(AvailabilityServiceContextV4 availabilityServiceContextV3,
                                                           com.genpt.nsight.v2.SplunkLogger splunkLogger,
                                                           ListenableFuture<InventoryResponseMultiProduct> gcpResponseFutures,
                                                           Map<String, List<com.genpt.nsight.v4.model.Part>> nsightInventoryMap){
        long invResponseProcessingStartTime = System.currentTimeMillis();
        availabilityServiceContextV3.getSitesWithInventory().clear();
        if(pilotSellingChannelsResponseFromGcp){// need the data from GCP in final response
            Map<String, List<com.genpt.nsight.v4.model.Part>> gcpInventoryMap =
                    asyncInventoryComparator.getGCPInventoryMultiProduct(gcpResponseFutures, availabilityServiceContextV3.getRequest());
            if(gcpInventoryMap==null)
                gcpInventoryMap = new HashMap<>();
            availabilityServiceContextV3.setGcpInventory(gcpInventoryMap);
            availabilityServiceContextV3.setInventory(gcpInventoryMap);
            List<List<com.genpt.nsight.v4.model.Part>> list = gcpInventoryMap.values().stream().map(parts -> parts.stream().collect(Collectors.toList())).collect(Collectors.toList());
            List<com.genpt.nsight.v4.model.Part> flat = list.stream()
                    .flatMap(List::stream)
                    .collect(Collectors.toList());
            availabilityServiceContextV3.setSiteWithPartsAvailable(flat.stream().map(part -> part.getSite().getSiteId()).collect(Collectors.toSet()));
            availabilityServiceContextV3.setSitesWithInventory(flat.stream().map(part -> part.getSite().getSiteId()).collect(Collectors.toSet()));

            LOGGER.info(splunkLogger.toString() + " Processed MultiProduct GCP Inventory Response for Configured Selling Channels in {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
            //at this point we can trigger an async request to do the compare
            // using nsightInventoryMap and gcpInventoryMap
            if(enableCompare){
                invResponseProcessingStartTime = System.currentTimeMillis();
                asyncInventoryComparator.compareResponsesAsync(ImmutableMap.copyOf(gcpInventoryMap),ImmutableMap.copyOf(nsightInventoryMap),new com.genpt.nsight.v1.SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest()));
                LOGGER.info(splunkLogger + " Started Only MultiProduct GCP/Nsight Response Comparing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
            }else
                LOGGER.info(splunkLogger + " Skipped MultiProduct GCP/Nsight Response Comparing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        }else{//GCP responses only for async compare with NSight
            //only async processing of futures and compare. no need to serve GCP in response
            invResponseProcessingStartTime = System.currentTimeMillis();
            asyncInventoryComparator.transfromGCPFuturesMultiProduct(ImmutableMap.copyOf(availabilityServiceContextV3.getInventory()), gcpResponseFutures, availabilityServiceContextV3.getRequest(), enableCompare);
            LOGGER.info(splunkLogger + " Started MultiProduct GCP/Nsight Response Comparing Processing Async Thread in  {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        }
    }

    @Override
    public com.genpt.nsight.v4.model.ProductAvailabilityResponse getProductAvailability(AvailabilityServiceContextV4 availabilityServiceContextV3, boolean isVerification) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());

        LOGGER.debug(splunkLogger.toString()+ " request payload{} ", availabilityServiceContextV3.getRequest().toString() );

        //if password is passed in for OtherSites
        if (availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().getOtherSites() != null){
            for (SiteIdentifier siteIdentifier : availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().getOtherSites().getSiteIdentifiers()){
                if (siteIdentifier.getPassword() != null && siteIdentifier.getName()!=null) {
                    //get the siteID
                    String siteID = nSightCache.getAliasToSiteIdMap().get(siteIdentifier.getName().concat("|").concat(siteIdentifier.getValue()));
                    if (siteID != null && siteID.trim().length() > 0)
                        availabilityServiceContextV3.getOtherSitesPassword().put(siteID, siteIdentifier.getPassword());
                    availabilityServiceContextV3.getOtherSitesStoreNumber().put(siteID, siteIdentifier.getValue());

                }
            }
        }

        long startTime = System.currentTimeMillis();
/*        final ZonedDateTime now = ZonedDateTime.now(ZoneId.of("UTC"));
        if(availabilityServiceContextV3.getTimeOfDay() == 0L)
        {
            availabilityServiceContextV3.setDayOfWeekLocal(now.getDayOfWeek().getValue() - 1);
            availabilityServiceContextV3.setTimeOfDay(getMillisFromMidnight(now));
        }
        final ZonedDateTime beginningOfDay = now.truncatedTo(ChronoUnit.DAYS);*/

        ProductAvailabilityRequest request = availabilityServiceContextV3.getRequest();

        Requestor requestor = request.getHeaderRequest().getRequestor();
        String siteId = nSightCache.getAliasToSiteIdMap().get(requestor.getRequestorIdentifier().getType() + "|" + requestor.getRequestorIdentifier().getValue());
        //Since this is for verification, even inactive sites should be allowed
        if (siteId ==null && isVerification) {
            siteId = nSightCache.getInactiveAliasToSiteIdMap().get(requestor.getRequestorIdentifier().getType() + "|" + requestor.getRequestorIdentifier().getValue());
        }

        SiteDTO callingStoreDTO = nSightCache.getSiteIdToSiteMap().get(siteId);
        if(availabilityServiceContextV3.getTimeOfDay() == 0L && !isVerification) {
            availabilityServiceContextV3.setDayOfWeekLocalStore(callingStoreDTO);
        }

        //Set Requesting Site
        availabilityServiceContextV3.setRequestingSite(nSightCache.getSiteIdToSiteMap().get(siteId));

        //------ Needed for V4 Aggregate services
        //need to get the local time of the store
        if (!isVerification) {
            String zoneID = callingStoreDTO.getTimeZoneID();
            ZonedDateTime localNow = ZonedDateTime.now(TimeZone.getTimeZone(zoneID).toZoneId());
            availabilityServiceContextV3.setStoreCurrentTime(localNow);
            availabilityServiceContextV3.setStoreDayOfWeek(localNow.getDayOfWeek().name());
            //set the store hours along with the other calculated times
            ArrayList<StoreHours> storeHoursList = nSightCache.getStoreHoursList().get(siteId);//map for each day of the week
            if (storeHoursList != null) {
                availabilityServiceContextV3.setStoreOpenCloseHours(storeHoursList, localNow, zoneID);
            } else {
                //do not have store hours
                LOGGER.warn(splunkLogger.toString() + "No Store hours for Site ID  {} ", siteId);
            }
        }

        //also set the time based on the timezone from site table
        if (callingStoreDTO!= null && callingStoreDTO.getTimeZone()!=null) {
            try {
                availabilityServiceContextV3.initializeStoreTimeZoneValues(callingStoreDTO.getTimeZone());
            }catch(Exception ex){
                LOGGER.warn(splunkLogger.toString()+"TimeZone Parsing Exception  {} ", ex.toString());
            }
        }else
            LOGGER.warn(splunkLogger.toString()+ " No TimeZone Available for Site  {} ", siteId);

        //Make List of Parts to be queried
        inventoryServiceV2.preparePartList(availabilityServiceContextV3);

        //Identify Related Sites
        long relatedSiteStartTime = System.currentTimeMillis();
        if (availabilityServiceContextV3.getRequestingSite().getSiteType().equalsIgnoreCase("Store"))
            siteRelationshipServiceV3.setPrimaryDC(availabilityServiceContextV3);

        siteRelationshipServiceV3.setRelatedSites(availabilityServiceContextV3);
        LOGGER.debug(splunkLogger.toString() + " Time to prepare relatedSiteInfo {} ms", (System.currentTimeMillis() - relatedSiteStartTime));

        //Set Inventory Futures for NONNapa
        long queryStartTime = System.currentTimeMillis();
        if (availabilityServiceContextV3.isCallNonTAMS())
        {
            inventoryServiceV2.setNonNapaInventoryFutures(availabilityServiceContextV3, nSightCache.getSiteIdToSiteMap());
            LOGGER.debug(splunkLogger.toString()+" Submitted NonNapa Inventory queries in {} ms", (System.currentTimeMillis() - queryStartTime));
        }
        // lets set the boolean whether free delivery calls go to GCP
        availabilityServiceContextV3.setFreeDeliveryGCPRequestNeeded(deliveryEstimateService.needGCPFreeDeliveryCall(availabilityServiceContextV3));
        // make these calls prior to inventory calls. the check is done inside the function
        callFreeDeliveryFromGCP(isVerification,availabilityServiceContextV3,splunkLogger);
        // make the calls for inventory
        handleAvailabilityCalls(availabilityServiceContextV3, splunkLogger);
        long invResponseProcessingStartTime;

        //let call the delivery data after we have got the inventory
        //Request DeliveryEstimates. We will also need to handle scenarios where S2S and S2Z are required
        if (!isVerification) {
            if (availabilityServiceContextV3.isNeedSiteToSiteAndZip()) {
                if (availabilityServiceContextV3.getZipCodesSet().size() > 0) {
                    long deliveryEstimateQueryStartTime = System.currentTimeMillis();
                    deliveryEstimateServiceByZip.setDeliveryEstimateFutures(availabilityServiceContextV3);
                    LOGGER.debug(splunkLogger.toString() + " Submit SiteToZip RawDelivery Estimate queries in {} ms", (System.currentTimeMillis() - deliveryEstimateQueryStartTime));

                }
                long deliveryEstimateQueryStartTime = System.currentTimeMillis();
                deliveryEstimateService.setDeliveryEstimateFutures(availabilityServiceContextV3);
                LOGGER.debug(splunkLogger.toString() + "  Submit SitetoSite RawDelivery Estimate queries in {} ms", (System.currentTimeMillis() - deliveryEstimateQueryStartTime));
            } else {
                //if zipcode is present use that
                if (availabilityServiceContextV3.getZipCodesSet().size() > 0) {
                    long deliveryEstimateQueryStartTime = System.currentTimeMillis();
                    deliveryEstimateServiceByZip.setDeliveryEstimateFutures(availabilityServiceContextV3);
                    LOGGER.debug(splunkLogger.toString() + " Submit SiteToZip RawDelivery Estimate queries in {} ms", (System.currentTimeMillis() - deliveryEstimateQueryStartTime));

                } else {
                    long deliveryEstimateQueryStartTime = System.currentTimeMillis();
                    deliveryEstimateService.setDeliveryEstimateFutures(availabilityServiceContextV3);
                    LOGGER.debug(splunkLogger.toString() + "  Submit SitetoSite RawDelivery Estimate queries in {} ms", (System.currentTimeMillis() - deliveryEstimateQueryStartTime));
                }
            }
        }

        invResponseProcessingStartTime = System.currentTimeMillis();
        if (availabilityServiceContextV3.isCallNonTAMS()) {
            inventoryServiceV2.processNonNapaInventoryFutures(availabilityServiceContextV3);
            LOGGER.info(splunkLogger.toString()+ " Processed TAMS/RPM Inventory Query Response in {} ms", (System.currentTimeMillis() - invResponseProcessingStartTime));
        }

        //Now call lead times futures only after related sites are setup
        leadTimesServiceV3.setLeadTimesFutures(availabilityServiceContextV3);

        //process lead times futures
        leadTimesServiceV3.setLeadTimesBySite(availabilityServiceContextV3);

        //process the rawDelivery estimate futures
        Stopwatch sw = Stopwatch.createStarted();
        if (!isVerification) {
            if (availabilityServiceContextV3.isNeedSiteToSiteAndZip()) {
                if (availabilityServiceContextV3.getDeliveryEstimateFutures() != null)
                    deliveryEstimateService.setDeliveryBySite(availabilityServiceContextV3);
                if (availabilityServiceContextV3.getZipCodesSet().size() > 0) {
                    deliveryEstimateServiceByZip.setDeliveryByZip(availabilityServiceContextV3);
                }
            } else {
                if (availabilityServiceContextV3.getDeliveryEstimateFutures() != null || availabilityServiceContextV3.getDeliveryZipEstimateFutures() != null) {
                    if (availabilityServiceContextV3.getZipCodesSet().size() > 0) {
                        deliveryEstimateServiceByZip.setDeliveryByZip(availabilityServiceContextV3);
                    } else {
                        deliveryEstimateService.setDeliveryBySite(availabilityServiceContextV3);
                    }
                }
            }
            LOGGER.info(splunkLogger.toString() + " Processed All Free and Paid Delivery Responses in {}", sw.stop());
        }

        //Merge the Napa and non Napa inventory MAPs based on partnumber+lineabbrev key
        if (availabilityServiceContextV3.getNonNapaInventory()!= null && availabilityServiceContextV3.getNonNapaInventory().size() > 0 ){//iterate non napa
            for ( Map.Entry<String, List<com.genpt.nsight.v4.model.Part>> NonNapaentry : availabilityServiceContextV3.getNonNapaInventory().entrySet() ){
                if (availabilityServiceContextV3.getInventory().containsKey(NonNapaentry.getKey())){//non napa has the same key
                    //check if we need to populate pricing data
                    //we have the pricing results from non napa
                    List<com.genpt.nsight.v4.model.Part>  nonNapaEntryValue = NonNapaentry.getValue();// this is a list of sites from POS for that part
                    //iterate this list to find the price from the call store response
                    //we should also get the quantity from the POS response
                    HashSet<com.genpt.nsight.v4.model.Part> mergedSitesSet =  new HashSet<>();// this is sites that are in Cassandra and POS for this part
                    for (com.genpt.nsight.v4.model.Part nonNapaPart : nonNapaEntryValue){//iterate over non napa sites for a key(part)
                        //check if we got success from POS and copy it over. if failiure, we also need to get the failiure reason
                        if (nonNapaPart.getPrice()!=null || (nonNapaPart.getSite() !=null && nonNapaPart.getSite().getSuperSedes()!=null) && nonNapaPart.getSite().getSuperSedes().size()>0
                                || (nonNapaPart.getSite() !=null && nonNapaPart.getSite().getQuantity()!=null) || nonNapaPart.getSite().getTracking().getIsSuccess()==false){
                            //get the corresponding entry from the Cassandra resultset
                            List<com.genpt.nsight.v4.model.Part>  nsightEntryValue = availabilityServiceContextV3.getInventory().get(NonNapaentry.getKey());
                            for (com.genpt.nsight.v4.model.Part nsightPart : nsightEntryValue){//this is the part with list of sites, iterate each site
                                if (nonNapaPart.getSite().getSiteId().compareToIgnoreCase(nsightPart.getSite().getSiteId()) == 0) {// this site has the part in cassandra
                                    //take the price, supersedes and quantity from the POS response
                                    //Price and Supersedes are at the Part level and should be from Local Store
                                    if (nonNapaPart.getSite().getSiteId().compareToIgnoreCase(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier())==0) {
                                        if (nonNapaPart.getPrice() != null)
                                            nsightPart.setPrice(nonNapaPart.getPrice());
                                        if (nonNapaPart.getSite().getSuperSedes() != null)
                                            nsightPart.getSite().setSuperSedes(nonNapaPart.getSite().getSuperSedes());
                                    }
                                    if (nonNapaPart.getSite().getQuantity()!=null) {
                                        //we need to get the individual attributes. Only available from POS. rest from Snight
                                        //nsightPart.getSite().setQuantity(part.getSite().getQuantity());
                                        nsightPart.getSite().getQuantity().setAvailable(nonNapaPart.getSite().getQuantity().getAvailable());
                                        nsightPart.getSite().getQuantity().setOnHand(nonNapaPart.getSite().getQuantity().getOnHand());
                                        nsightPart.getSite().getQuantity().setWorkInProgress(nonNapaPart.getSite().getQuantity().getWorkInProgress());
                                        nsightPart.getSite().setAvailableQuantity(nonNapaPart.getSite().getAvailableQuantity());
                                    }
                                    if (nonNapaPart.getSite().getPartAttributes() != null) {
                                        nsightPart.getSite().getPartAttributes().setCurrentClass(nonNapaPart.getSite().getPartAttributes().getCurrentClass());
                                        nsightPart.getSite().getPartAttributes().setGroupCode(nonNapaPart.getSite().getPartAttributes().getGroupCode());
                                        nsightPart.getSite().getPartAttributes().setPartDescription(nonNapaPart.getSite().getPartAttributes().getPartDescription());
                                    }
                                    //need to remove this site from the nonnapainventory map. Store that in a list and remove them at the end
                                    mergedSitesSet.add(nonNapaPart);
                                }//matched a site for this part from cassandra
                            }//iterate each site where this part is in cassandra
                        }//non napa part has valid data
                    }//iterate over non napa sites for this part
                    for (com.genpt.nsight.v4.model.Part mergedSite : mergedSitesSet)
                        nonNapaEntryValue.removeIf(nonNapaPart -> mergedSite.getSite().getSiteId()==nonNapaPart.getSite().getSiteId() );
                    if (nonNapaEntryValue.size() > 0)
                        availabilityServiceContextV3.getInventory().get(NonNapaentry.getKey()).addAll(NonNapaentry.getValue());
                }else{ //doesnt exist in Main Map
                    // only if its not a virtual kit, we just need the price and not availability(will be 0 from store)
                    String[] VKKeyArray = NonNapaentry.getKey().split("-");
                    String VKKey = VKKeyArray[1]+"-"+VKKeyArray[0];
                    if (availabilityServiceContextV3.getParentPartMap() !=null && availabilityServiceContextV3.getParentPartMap().containsKey(VKKey)){
                        //its a virtual kit just get the price and add it to the KIT info
                        com.genpt.nsight.v4.model.Part vkPart = availabilityServiceContextV3.getParentPartMap().get(VKKey);
                        List<com.genpt.nsight.v4.model.Part>  pricingEntry = NonNapaentry.getValue();
                        for (com.genpt.nsight.v4.model.Part part : pricingEntry) {
                            if (part.getPrice()!=null) {
                                vkPart.setPrice(part.getPrice());
                                break;
                            }
                            if (part.getTracking()!=null){
                                vkPart.setTracking(part.getTracking());
                                break;
                            }
                        }
                    }else//not a virtual kit, is eligible to be merged
                        availabilityServiceContextV3.getInventory().put(NonNapaentry.getKey(),NonNapaentry.getValue());
                }
            }
        }

        //Construct the Detail response by iterating over the parts
        long buildResponseStartTime = System.currentTimeMillis();

        com.genpt.nsight.v4.model.DetailResponse detailResponse = new com.genpt.nsight.v4.model.DetailResponse();
        if (availabilityServiceContextV3.getRequest().getDetailRequest().getShippingDestination()!=null){
            detailResponse.setShippingDestination(availabilityServiceContextV3.getRequest().getDetailRequest().getShippingDestination());
        }// set the shipping destinations from the request

        detailResponse.setParts(new ArrayList<>());
        //get the list of parts that came in the request
        final Set<String> partsInRequest = new HashSet<>(availabilityServiceContextV3.getInputParts().keySet());

        availabilityServiceContextV3.getInventory().forEach((k, v) -> {
            com.genpt.nsight.v4.model.Part part = new com.genpt.nsight.v4.model.Part();
            final BigDecimal[] combinedAvailability = new BigDecimal[1];
            combinedAvailability[0]= new BigDecimal(0);

            v.forEach(p -> {
                //Virtual Kit
                KitComponent kitComponent = null;
                String fullPartNumber = p.getLineAbbrev() + "-" + p.getPartNumber();

                if (part.getPartNumber() == null) {
                    //set the tracking for part if its not a success
                    if (p.getTracking()!= null && !p.getTracking().getIsSuccess()){
                        part.setTracking(p.getTracking());
                    }
                    part.setPartNumber(p.getPartNumber());
                    part.setLineAbbrev(p.getLineAbbrev());
                    //partsInRequest.remove(fullPartNumber); //This part exists in Input as well as output
                    part.setSites(new HashSet<>(v.size()));

                    //Catalog PartId
                    com.genpt.nsight.v4.model.Part inputPart = availabilityServiceContextV3.getInputParts().get(fullPartNumber);
                    if (inputPart != null) {
                        part.setCatalogPartId(inputPart.getCatalogPartId());
                        part.setSequenceNumber(inputPart.getSequenceNumber());
                    }
                }

                if(p.getMessage()!=null) {
                    part.setMessage(p.getMessage());
                }
                //set pricing only if its the calling store
                if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedPricing() ){
                    //pricing data wont be available for suppliers and Dcs
                    if (availabilityServiceContextV3.getRequestingSite().getSiteType().compareToIgnoreCase("DC")==0){
                        Tracking dcTracking = new Tracking();
                        dcTracking.setIsSuccess(false);
                        dcTracking.setLog(priceUnavailableDC);
                        PriceType priceType = new PriceType();
                        priceType.setTracking(dcTracking);
                        part.setPrice(priceType);
                    }
                    else if (availabilityServiceContextV3.getRequestingSite().getSiteType().compareToIgnoreCase("Supplier")==0){
                        Tracking dcTracking = new Tracking();
                        dcTracking.setIsSuccess(false);
                        dcTracking.setLog(priceUnavailableSupplier);
                        PriceType priceType = new PriceType();
                        priceType.setTracking(dcTracking);
                        part.setPrice(priceType);
                    }
                    else if (availabilityServiceContextV3.getRequestingSite().getSiteIdentifier().compareToIgnoreCase(p.getSite().getSiteId()) == 0 && availabilityServiceContextV3.getRequestingSite().getSiteType().compareToIgnoreCase("Store")==0){
                        if (p.getPrice() != null)
                            part.setPrice(p.getPrice());
                        else{
                            //if we are not able to get pricing data for the calling store , indicate that in the Tracking
                            PriceType priceType = new PriceType();
                            Tracking storeTracking = new Tracking();
                            storeTracking.setIsSuccess(false);
                            storeTracking.setLog(Collections.singletonList(new Log("2001", "Unavailable to Retrieve Pricing Data from " + availabilityServiceContextV3.getRequestingSite().getPosType())));
                            if(availabilityServiceContextV3.getPricingStatusCode() != null && availabilityServiceContextV3.getPricingStatusMessage() != null) {
                                storeTracking.setFault(Fault.builder().code(availabilityServiceContextV3.getPricingStatusCode()).message(availabilityServiceContextV3.getPricingStatusMessage()).build());
                            }
                            priceType.setTracking(storeTracking);
                            part.setPrice(priceType);
                        }
                    }
                }
                if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedSuperSedes() ) {
                    //check if we have supersedes
                    if (p.getSite().getSuperSedes()!=null){
                        part.setSuperSedes(p.getSite().getSuperSedes());
                    }
                }

                com.genpt.nsight.v4.model.Site site = (com.genpt.nsight.v4.model.Site) p.getSite();

                //we will populate this site only when its active for all requestors except JDE and TAMS
                if (site != null) {
                    SiteDTO siteDTO = nSightCache.getSiteIdToSiteMap().get(site.getSiteId());

                    if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedCityStateInfo() != null && availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedCityStateInfo()) {
                        site.setSiteCity(siteDTO.getCity());
                        site.setSiteState(siteDTO.getState());
                    }

                    if (siteDTO.getIsActive() || (availabilityServiceContextV3.getApplicationID().equalsIgnoreCase("JDE") ||
                            availabilityServiceContextV3.getApplicationID().equalsIgnoreCase("TAMS") || availabilityServiceContextV3.getApplicationID().equalsIgnoreCase("ISSCQUERY"))) {

                        //Populate RawDelivery Estimate
                        if (availabilityServiceContextV3.isNeedSiteToSiteAndZip()){// this would be for aggregate service that calls the freepaid v4
                            if (availabilityServiceContextV3.getDeliveryBySite() != null && availabilityServiceContextV3.getDeliveryBySite().size() > 0) {
                                site.setRawDelivery(availabilityServiceContextV3.getDeliveryBySite().get(site.getSiteId()));
                            }
                        }
                        else {// need data from either sitetosite or sitetozip
                            if (availabilityServiceContextV3.getDeliveryBySite() != null && availabilityServiceContextV3.getDeliveryBySite().size() > 0) {
                                site.setRawDelivery(availabilityServiceContextV3.getDeliveryBySite().get(site.getSiteId()));
                            }
                            //if the data requested was from Sitetozip
                            if (availabilityServiceContextV3.getDeliveryByZip() != null && availabilityServiceContextV3.getDeliveryByZip().size() > 0) {
                                site.setRawDelivery(availabilityServiceContextV3.getDeliveryByZip().get(site.getSiteId()));
                            }
                        }

                        List<SiteRelationship> siteRelationshipList = availabilityServiceContextV3.getSiteRelationMap().get(site.getSiteId());

                        if (siteRelationshipList == null) {
                            if (nSightCache.getDcSet().contains(site.getSiteId())) {
                                siteRelationshipList = Collections.singletonList(new SiteRelationship(null, "AllOtherDCs"));
                            } else if (nSightCache.getSupplierIdSet().contains(site.getSiteId())) {
                                siteRelationshipList = Collections.singletonList(new SiteRelationship(null, "AllSuppliers"));
                            }
                            if (siteRelationshipList != null) {
                                availabilityServiceContextV3.getSiteRelationMap().put(site.getSiteId(), siteRelationshipList);
                            }
                        }

                        if (siteRelationshipList != null && siteRelationshipList.size() > 0) {
                            site.setSiteRelationships(siteRelationshipList);
                            // also check for combined inventory
                            boolean isCombinedFriend =  siteRelationshipList.stream().filter(o -> o.getSiteRelationshipType().equals("Combined Inventory Friends")).findFirst().isPresent();
                            boolean isLocal =  siteRelationshipList.stream().filter(o -> o.getSiteRelationshipType().equals("LocalStore")).findFirst().isPresent();

                            if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedCombinedInventory())
                                if(isCombinedFriend || isLocal) {
                                    if (site.getQuantity()!=null)
                                        combinedAvailability[0] = combinedAvailability[0].add( site.getQuantity().getAvailable());
                                }
                        }

                        com.genpt.nsight.v4.model.Part parentPart = null;
                        if (availabilityServiceContextV3.getVirtualPart2ParentPartMap() != null) {
                            Set<String> parentPartKeys = availabilityServiceContextV3.getVirtualPart2ParentPartMap().get(fullPartNumber);
                            if (parentPartKeys != null && parentPartKeys.size() > 0) {
                                for (String parentPartKey : parentPartKeys) {
                                    if (parentPartKey != null) {
                                        partsInRequest.remove(parentPartKey); //This is parent part that came in Input
                                        parentPart = availabilityServiceContextV3.getParentPartMap().get(parentPartKey); //get the ParentPart - the one that has come in request
                                        kitComponent = availabilityServiceContextV3.getKitComponentMap().get(fullPartNumber); //Get the kitComponent for this Part from response
                                        //if pricing was requested for VK, populate tracking if unable to retrieve price
                                        if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedPricing() && parentPart.getTracking()!=null && parentPart.getTracking().getFault()!=null && parentPart.getPrice()==null){
                                            PriceType priceType = new PriceType();
                                            Tracking vkPriceTracking = new Tracking();
                                            vkPriceTracking.setIsSuccess(false);
                                            vkPriceTracking.setLog(Collections.singletonList(new Log("2001", "Unavailable to Retrieve Pricing Data from " + availabilityServiceContextV3.getRequestingSite().getPosType())));
                                            if(availabilityServiceContextV3.getPricingStatusCode() != null && availabilityServiceContextV3.getPricingStatusMessage() != null) {
                                                vkPriceTracking.setFault(Fault.builder().code(availabilityServiceContextV3.getPricingStatusCode()).message(availabilityServiceContextV3.getPricingStatusMessage()).build());
                                            }
                                            priceType.setTracking(vkPriceTracking);
                                            parentPart.setPrice(priceType);
                                        }

                                        parentPart.setTracking(tracking);
                                        //update parent part data
                                        //Parent part is present if this part was from a Virtual Kit
                                        if (parentPart != null) { //If Parent part is present
                                            //could be a part of more than one virtual kit
                                            if (parentPart.getSiteMap() == null) {
                                                parentPart.setSiteMap(new HashMap<>());
                                            }
                                            int parentPartAvailable = 0;
                                            if (availabilityServiceContextV3.isAggregate() )
                                                parentPartAvailable = site.getAvailableQuantity() ;
                                            else{
                                                if(site.getQuantity() !=null)
                                                    parentPartAvailable = site.getQuantity().getAvailable().intValue() / kitComponent.getQuantityRequired(); //Number of parts that can be constructed with the part
                                            }
                                            if (parentPartAvailable > 0) {    //Create Site for Parent Part only if Valid Quantity

                                                com.genpt.nsight.v4.model.Site parentPartSite = parentPart.getSiteMap().get(site.getSiteId());  //Get the Site identified by current SiteId
                                                // we also need to check the site for country logic, either the same country or the sites have a explicit relationship
                                                if ((availabilityServiceContextV3.getRequestingSite().getCountry()!= null
                                                        && nSightCache.getSiteIdToSiteMap().get(site.getSiteId()) !=null
                                                        && nSightCache.getSiteIdToSiteMap().get(site.getSiteId()).getCountry() != null
                                                        && availabilityServiceContextV3.getRequestingSite().getCountry().compareToIgnoreCase(nSightCache.getSiteIdToSiteMap().get(site.getSiteId()).getCountry())==0)
                                                        || (siteRelationshipList!=null && siteRelationshipList.size() > 1 )
                                                        || (siteRelationshipList!=null && siteRelationshipList.size() == 1 &&
                                                        (siteRelationshipList.get(0).getSiteRelationshipType().compareToIgnoreCase("AllOtherDCs") != 0) &&
                                                        (siteRelationshipList.get(0).getSiteRelationshipType().compareToIgnoreCase("AllSuppliers") != 0))
                                                ) {
                                                    if (parentPartSite == null) {   //If Site is null, create the copy of existing Site
                                                        parentPartSite = site.copy();
                                                        parentPart.getSiteMap().put(parentPartSite.getSiteId(), parentPartSite);
                                                        Quantity quantity = new Quantity();
                                                        quantity.setNumVirtualPartsUnsatisfied(parentPart.getProductKit().getKitComponents().size());
                                                        parentPartSite.setQuantity(quantity);
                                                        //since we dont have the below for Virtual Kits, we will have default values.
                                                        PartAttributes partAttributes = new PartAttributes();
                                                        partAttributes.setUnitOfMeasure("EA");
                                                        partAttributes.setPartDescription("Virtual Kit");
                                                        parentPartSite.setPartAttributes(partAttributes);
                                                    }

                                                    if (parentPartSite.getQuantity().getAvailable() == null || parentPartSite.getQuantity().getAvailable().intValue() > parentPartAvailable) {
                                                        parentPartSite.getQuantity().setAvailable(new BigDecimal(parentPartAvailable));
                                                        parentPartSite.setAvailableQuantity(parentPartAvailable);
                                                    }

                                                    parentPartSite.getQuantity().setNumVirtualPartsUnsatisfied(parentPartSite.getQuantity().getNumVirtualPartsUnsatisfied() - 1);
                                                }//endif country check
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        //update the individual parts site
                        //Check if site is canada or non
                        //if the calling site country and the site country is  the same unless a relationship is setup between the 2
                        if (availabilityServiceContextV3.getRequestingSite().getCountry()!= null
                                && nSightCache.getSiteIdToSiteMap().get(site.getSiteId()) !=null  && nSightCache.getSiteIdToSiteMap().get(site.getSiteId()).getCountry() != null&& availabilityServiceContextV3.getRequestingSite().getCountry().compareToIgnoreCase(nSightCache.getSiteIdToSiteMap().get(site.getSiteId()).getCountry())==0){
                            part.getSites().add(site);
                            partsInRequest.remove(fullPartNumber); //This part exists in Input as well as output
                        }else{
                            //we dont have country code or the country code is not a match
                            //if its a DC or a supplier we would still be interested in the site if a relationship is there between the two
                            //if its relationship is onlu AllOtherDcs or AllSuppliers then we can drop it
                            if (siteRelationshipList!=null && siteRelationshipList.size()==1
                                    && (siteRelationshipList.get(0).getSiteRelationshipType().compareToIgnoreCase("AllOtherDCs")==0
                                    || siteRelationshipList.get(0).getSiteRelationshipType().compareToIgnoreCase("AllSuppliers")==0))
                                LOGGER.debug(splunkLogger.toString() + " Dropping site from response {} ", site.getSiteId());
                            else{
                                part.getSites().add(site);
                                partsInRequest.remove(fullPartNumber); //This part exists in Input as well as output
                            }

                        }
                    }//site is active
                }//site check for null
            });
            //Need Inventory for Combined friends
            if (availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedCombinedInventory()){
                part.getSites().forEach(site -> {
                    if (site.getSiteRelationships().stream().filter(o -> o.getSiteRelationshipType().equals("LocalStore")).findFirst().isPresent()) {
                        if ( site.getQuantity()!= null )
                            site.getQuantity().setCombinedAvailability(new BigDecimal(0).add(combinedAvailability[0]));
                    }
                } );
            }
            //add part only if it has sites
            if (part.getSites() != null && part.getSites().size() > 0) {
                //Populate tracking at Part Level
                part.setTracking(tracking);
                detailResponse.getParts().add(part);
            }

        });//end of each part

        //Add all Parent Parts
        if (availabilityServiceContextV3.getParentPartMap() != null) {
            availabilityServiceContextV3.getParentPartMap().forEach((partKey, part) -> {
                if (part.getSiteMap() != null && part.getSiteMap().size() > 0) {
                    Set<com.genpt.nsight.v4.model.Site> siteList = part.getSiteMap().values().stream().filter(site -> site.getQuantity().getNumVirtualPartsUnsatisfied() == 0).collect(Collectors.toSet());
                    if(siteList != null && siteList.size() > 0) {
                        part.setSites(siteList);
                    }
                    part.getSiteMap().clear();
                }
                detailResponse.getParts().add(part);
            });
        }

        com.genpt.nsight.v4.model.ProductAvailabilityResponse productAvailabilityResponse = availabilityServiceContextV3.getProductAvailabilityResponseTemplate(availabilityServiceContextV3);
        productAvailabilityResponse.setDetailResponse(detailResponse);
        productAvailabilityResponse.setTracking(tracking);

        //For missing Parts in response add Log
        if(partsInRequest.size() > 0) {
            if(partsInRequest.size() == availabilityServiceContextV3.getRequest().getDetailRequest().getParts().size()) { //All Parts are missing in Response
                productAvailabilityResponse.setTracking(new Tracking());
                productAvailabilityResponse.getTracking().setLog(noPartsFound);
            }

            if(productAvailabilityResponse.getDetailResponse().getParts() == null) {
                productAvailabilityResponse.getDetailResponse().setParts(new ArrayList<>());
            }

            partsInRequest.forEach(fullPartNumber -> {
                com.genpt.nsight.v4.model.Part part = availabilityServiceContextV3.getInputParts().get(fullPartNumber);
                if (part.getTracking() == null) {
                    part.setTracking(new Tracking());
                    part.getTracking().setLog(partNotFound);
                }
                if (availabilityServiceContextV3.getParentPartMap() == null || availabilityServiceContextV3.getParentPartMap().get(fullPartNumber) == null) { //This is not a virtual Part
                    productAvailabilityResponse.getDetailResponse().getParts().add(part);
                }
            });
        }
        int totalPartsRequested = getTotalPartsRequested(availabilityServiceContextV3);
        LOGGER.debug(splunkLogger.toString() + " Time taken to build response in ms = {} and the total parts requested are = {} ",  (System.currentTimeMillis() - buildResponseStartTime), totalPartsRequested);
        LOGGER.info(splunkLogger.toString()+ " Total time within Details MicroService  in = {} and the total parts requested are = {} " , (System.currentTimeMillis() - startTime), totalPartsRequested);

        return productAvailabilityResponse;
    }

    private void callFreeDeliveryFromGCP(boolean isVerification, AvailabilityServiceContextV4 availabilityServiceContextV3, SplunkLogger splunkLogger){
        if (!isVerification && availabilityServiceContextV3.isFreeDeliveryGCPRequestNeeded()){
            List<DeliveryService> deliveryServiceRequests = new ArrayList<>();
            for (Map.Entry<String,String> entry : availabilityServiceContextV3.getShipToSitesMap().entrySet()) {//for each destination
                String aliasValue = entry.getKey();
                DeliveryService deliveryService  = new DeliveryService();
                if(aliasValue!=null && !aliasValue.trim().isEmpty()){
                    deliveryService.setDestination(aliasValue);
                    deliveryService.setCorrelationId(availabilityServiceContextV3.getCorrelationID());
                    deliveryService.setClientId(availabilityServiceContextV3.getApplicationID());
                    deliveryServiceRequests.add(deliveryService);
                }
                //we would need to send the sources only if both the allSuppliers and allDCs are false
                if(!availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllOtherDCs()
                && !availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllSuppliers()){
                    //Now get the sources
                    deliveryService.setSources(availabilityServiceContextV3.getSourceLocationsToGCP());
                }
            }
            //make the call now
            deliveryEstimateService.setGCPFreeDeliveryFutures(deliveryServiceRequests,splunkLogger,availabilityServiceContextV3);
        }
    }

    private static int getTotalPartsRequested(AvailabilityServiceContextV4 availabilityServiceContextV4) {
        if(availabilityServiceContextV4 != null && availabilityServiceContextV4.getRequest() != null
                && availabilityServiceContextV4.getRequest().getDetailRequest() != null && availabilityServiceContextV4.getRequest().getDetailRequest().getParts() != null) {
            return availabilityServiceContextV4.getRequest().getDetailRequest().getParts().size();
        }
        return 0;
    }

    private void setSkipHQLineAbbrev(AvailabilityServiceContextV4 availabilityServiceContextV3) {
        InventoryRelationshipsRequested inventoryRelationshipsRequested = availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested();

        //check if we need to apply the hqabbrev logic
        if (!getNSightCache().getNSightCacheLoader().getNSightProperties().isUseHQAbbrevLogic()) {
            availabilityServiceContextV3.setSkipHQLineAbbrev(true);
        }else {//we need to apply hqabbrev logic
            if (availabilityServiceContextV3.getRequestingSite().getCountry() != null && availabilityServiceContextV3.getRequestingSite().getCountry().compareToIgnoreCase("CAN") == 0) {
                if (getNSightCache().getNSightCacheLoader().getNSightProperties().isApplyHQAbbrevUAP())
                    availabilityServiceContextV3.setSkipHQLineAbbrev(false);
                else {
                    final boolean skipHQLineAbbrev = (
                            inventoryRelationshipsRequested.isLocalStore() ||
                                    inventoryRelationshipsRequested.isServingDC()
                    ) &&
                            (
                                    !inventoryRelationshipsRequested.isAllOtherDCs() &&
                                            !inventoryRelationshipsRequested.isAllSuppliers() &&
                                            !inventoryRelationshipsRequested.isLocalStoreOwnershipGroup() &&
                                            (
                                                    inventoryRelationshipsRequested.getOtherRelationShips() == null ||
                                                            inventoryRelationshipsRequested.getOtherRelationShips().getRelationshipName() == null ||
                                                            inventoryRelationshipsRequested.getOtherRelationShips().getRelationshipName().size() == 0
                                            ) &&
                                            (
                                                    inventoryRelationshipsRequested.getOtherSites() == null ||
                                                            inventoryRelationshipsRequested.getOtherSites().getSiteIdentifiers() == null ||
                                                            inventoryRelationshipsRequested.getOtherSites().getSiteIdentifiers().size() == 0
                                            )
                            );

                    availabilityServiceContextV3.setSkipHQLineAbbrev(skipHQLineAbbrev);
                }
            } else {
                final boolean skipHQLineAbbrev = (
                        inventoryRelationshipsRequested.isLocalStore() ||
                                inventoryRelationshipsRequested.isServingDC()
                ) &&
                        (
                                !inventoryRelationshipsRequested.isAllOtherDCs() &&
                                        !inventoryRelationshipsRequested.isAllSuppliers() &&
                                        !inventoryRelationshipsRequested.isLocalStoreOwnershipGroup() &&
                                        (
                                                inventoryRelationshipsRequested.getOtherRelationShips() == null ||
                                                        inventoryRelationshipsRequested.getOtherRelationShips().getRelationshipName() == null ||
                                                        inventoryRelationshipsRequested.getOtherRelationShips().getRelationshipName().size() == 0
                                        ) &&
                                        (
                                                inventoryRelationshipsRequested.getOtherSites() == null ||
                                                        inventoryRelationshipsRequested.getOtherSites().getSiteIdentifiers() == null ||
                                                        inventoryRelationshipsRequested.getOtherSites().getSiteIdentifiers().size() == 0
                                        )
                        );

                availabilityServiceContextV3.setSkipHQLineAbbrev(skipHQLineAbbrev);
            }
        }
    }

    protected long getMillisFromMidnight(ZonedDateTime now) {
        ZonedDateTime midnight = now.truncatedTo(ChronoUnit.DAYS);
        Duration duration = Duration.between(midnight, now);
        TimeZone tz  =  TimeZone.getDefault();

        boolean isOn = ZoneId.of(tz.getID())  // Represent a specific time zone, the history of past, present, and future changes to the offset-from-UTC used by the people of a certain region.
                .getRules()                // Obtain the list of those changes in offset.
                .isDaylightSavings(        // See if the people of this region are observing Daylight Saving Time at a specific moment.
                        Instant.now()          // Specify the moment. Here we capture the current moment at runtime.
                )    ;                      // Returns a boolean.

        if (isOn)
            return duration.getSeconds() * 1000L +tz.getDSTSavings();
        else
            return duration.getSeconds()  * 1000L ;
    }

    private String getDateTime(int days, long timeInMills, ZonedDateTime now) {
        long minutes = timeInMills / (60 * 1000L);
        String dateTime = null;
        try {
            dateTime = DateTimeFormatter.ISO_INSTANT.format(now.plusDays(days).plusMinutes(minutes));

        } catch (Exception e) {
            e.printStackTrace();
        }
        return dateTime;
    }

}
