package com.genpt.nsight.v4.service.impl;

import com.datastax.driver.core.*;
import com.genpt.nsight.NSightCache;
import com.genpt.nsight.config.DeliveryTypes;
import com.genpt.nsight.config.NSightProperties;
import com.genpt.nsight.v1.model.*;
import com.genpt.nsight.v1.model.dto.SiteDTO;
import com.genpt.nsight.v1.model.serializer.DeliveryCutOffSerializer;
import com.genpt.nsight.v2.SplunkLogger;
import com.genpt.nsight.v4.AvailabilityServiceContextV4;
import com.genpt.nsight.v4.model.Delivery;
import com.genpt.nsight.v4.model.DeliveryEstimateFreePaid;
import com.genpt.nsight.v4.model.Expectation;
import com.genpt.nsight.v4.model.StoreHours;
import com.genpt.nsight.v4.service.Site2SiteDeliveryEstimateServiceV4;
import com.genpt.nsight.v4.service.delivery.AsyncDeliveryClient;
import com.genpt.nsight.v4.service.delivery.DeliveryOption;
import com.genpt.nsight.v4.service.delivery.DeliveryResponseV1;
import com.genpt.nsight.v4.service.delivery.DeliveryService;
import com.google.common.base.Stopwatch;
import com.google.common.collect.ImmutableList;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.SettableFuture;
import org.apache.commons.lang3.text.WordUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.time.*;
import java.time.Duration;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.stream.Collectors;

import static java.time.temporal.ChronoUnit.MILLIS;

@Service("Site2SiteDeliveryEstimateServiceV4")
public class Site2SiteDeliveryEstimateServiceV4Impl implements Site2SiteDeliveryEstimateServiceV4 {
    private String[] dayOfWeekColNames = new String[]{"monday_cutoff_time", "tuesday_cutoff_time", "wednesday_cutoff_time", "thursday_cutoff_time", "friday_cutoff_time", "saturday_cutoff_time", "sunday_cutoff_time"};
    public String[] leadTimesDayOfWeekColNames = new String[]{"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"};

    private static final Logger LOGGER = LoggerFactory.getLogger(Site2SiteDeliveryEstimateServiceV4Impl.class);

    @Autowired
    DeliveryTypes deliveryTypes;

    @Autowired
    private Session session;

    @Autowired
    NSightCache nSightCache;

    @Autowired
    NSightProperties nSightProperties;

    PreparedStatement deliveryEstimatePS;

    PreparedStatement deliveryEstimateFreePS;

    PreparedStatement deliveryEstimateFreeOptimizedPS;

    @Autowired
    AsyncDeliveryClient asyncDeliveryClient;
    private static long deliveryServiceTimeout;
    @Value("${delivery.deliveryServiceTimeout}")
    public void setDeliveryTimeout(long deliveryServiceTimeout){
        this.deliveryServiceTimeout=deliveryServiceTimeout;
    }
    @Value("${delivery.pilotSellingChannelsCallsToGcp}")
    private boolean pilotSellingChannelsCallsToGcp;
    @Value("${delivery.allFreeCallsToGcp}")
    private boolean allFreeCallsToGcp;
    @Value("${delivery.allFreeCallsToNsight}")
    private boolean allFreeCallsToNsight;
    @Value("${delivery.freeDeliveryPilotSellingChannels:123}")
    private String freeDeliveryPilotSellingChannels;
    @Value("${delivery.enablePublishFreeDelivery}")
    private boolean enablePublishFreeDelivery;

    @Autowired(required = false)
    private AsyncDeliveryPubsubPublisher asyncDeliveryPubsubPublisher;

    @PostConstruct
    public void init() {
        deliveryEstimatePS = session.prepare("SELECT delivery_service_zone, ship_to_site_identifier, delivery_service_type, ship_from_site_identifier, monday_cutoff_time, tuesday_cutoff_time, wednesday_cutoff_time, thursday_cutoff_time, friday_cutoff_time, saturday_cutoff_time, sunday_cutoff_time, service_duration_designator, service_duration_max_value, service_duration_min_value FROM site_to_site_delivery_matrix where ship_to_site_identifier = ? and delivery_service_type = ? and ship_from_site_identifier in ?;");
        deliveryEstimatePS.setConsistencyLevel(ConsistencyLevel.LOCAL_QUORUM);

        deliveryEstimateFreePS = session.prepare("SELECT ship_to_site_identifier, delivery_service_type, ship_from_site_identifier, day, cutoff, earliest_time, latest_time, delivery_display_name, service_duration_designator, service_duration_max_value, service_duration_min_value , delivery_category FROM site_to_site_free_delivery_by_day where ship_to_site_identifier = ? and ship_from_site_identifier in ?;");
        deliveryEstimateFreePS.setConsistencyLevel(ConsistencyLevel.LOCAL_QUORUM);

        deliveryEstimateFreeOptimizedPS = session.prepare("SELECT ship_to_site_identifier, delivery_service_type, ship_from_site_identifier, day, cutoff, earliest_time, latest_time, delivery_display_name, service_duration_designator, service_duration_max_value, service_duration_min_value , delivery_category FROM site_to_site_free_delivery_by_day where ship_to_site_identifier = ? and ship_from_site_identifier = ?;");
        deliveryEstimateFreeOptimizedPS.setConsistencyLevel(ConsistencyLevel.LOCAL_QUORUM);
    }



    @Override
    public ResultSetFuture getDeliveryEstimates(UUID toSiteId, String deliveryType, List<UUID> fromSiteList) {
        BoundStatement boundStatement = deliveryEstimatePS.bind(toSiteId,deliveryType,fromSiteList);
        return session.executeAsync(boundStatement);
    }

    @Override
    public ResultSetFuture getFreeDeliveryEstimates(UUID toSiteId, List<UUID> fromSiteList ,SplunkLogger splunkLogger) {
        BoundStatement boundStatement = deliveryEstimateFreePS.bind(toSiteId,fromSiteList);
        LOGGER.info(splunkLogger.toString() + " FreedeliveryByDay execute prepared statement To SiteID {}, FromSite ID: {} ", toSiteId, fromSiteList.toString());
        return session.executeAsync(boundStatement);
    }

    @Override
    public ResultSetFuture getFreeDeliveryEstimatesOptimized(UUID toSiteId, UUID fromSiteId,SplunkLogger splunkLogger) {
        BoundStatement boundStatement = deliveryEstimateFreeOptimizedPS.bind(toSiteId,fromSiteId);
        return session.executeAsync(boundStatement);
    }

    /// prepare all the cassandra query futures
    @Override
    public void setDeliveryEstimateFutures(AvailabilityServiceContextV4 availabilityServiceContextV3) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());
        //if requested only primary DC , just add that
        if (availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllOtherDCs())
            availabilityServiceContextV3.getFromSiteIds().addAll(nSightCache.getDcSet());
        else if (availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isServingDC()) {
         if (availabilityServiceContextV3.getPrimaryDC()!=null)
            availabilityServiceContextV3.getFromSiteIds().add(availabilityServiceContextV3.getPrimaryDC().getSiteId());
        }

        if (availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllSuppliers() )
            availabilityServiceContextV3.getFromSiteIds().addAll(nSightCache.getSupplierIdSet());

        // use the from sites as the list of related sites to fetch delivery data
/*
        List<UUID> fromSiteIdList = new ArrayList<>(availabilityServiceContextV3.getFromSiteIds().size());
        availabilityServiceContextV3.getFromSiteIds().forEach(siteId -> fromSiteIdList.add(UUID.fromString(siteId)));
*/
        //we will use the sites where we have found inventory instead. This is possible only if the inventory responses
        // ar processed prior to making these queries.
        List<UUID> fromSiteIdList = new ArrayList<>(availabilityServiceContextV3.getSitesWithInventory().size());
        availabilityServiceContextV3.getSitesWithInventory().forEach(siteId -> fromSiteIdList.add(UUID.fromString(siteId)));
        UUID fromSiteId = UUID.fromString(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier());
        //ad to map
        String requestorKey= availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getType()+"|"+ availabilityServiceContextV3.getRequest().getHeaderRequest().getRequestor().getRequestorIdentifier().getValue();
        availabilityServiceContextV3.getReverseShipToSitesMap().put(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier(),requestorKey);

        List<ResultSetFuture> deliveryTimeFutureResultSetList = new ArrayList<>();
        List<ResultSetFuture> deliveryTimeFreeFutureResultSetList = new ArrayList<>();

        //if there are entries in additional ship to sites, add those
        // it should at least have the requesting store
        for (Map.Entry<String,String> entry : availabilityServiceContextV3.getShipToSitesMap().entrySet()){
            String aliasName = entry.getValue();
            String aliasValue = entry.getKey();
            String fullAliasKey= aliasName + "|" + aliasValue;
            String shipToSiteID = nSightCache.getAliasToSiteIdMap().get(fullAliasKey);
            if (shipToSiteID!=null && shipToSiteID.trim().length()>0) {
                //need the following on the response path
                //add to the map
                availabilityServiceContextV3.getReverseShipToSitesMap().put(shipToSiteID, fullAliasKey);
                //for each rawDelivery type
                deliveryTypes.getDeliveryTypes().stream().map(deliveryType -> {
                    //we need only paid MOTS for this
                    if (nSightProperties.getPaidMOTS().contains(deliveryType))
                        return getDeliveryEstimates(UUID.fromString(shipToSiteID), deliveryType, fromSiteIdList);
                    else return null;// will this create a problem?
                }).forEach(deliveryTimeFutureResultSetList::add);

                //remove any nulls
                deliveryTimeFutureResultSetList.removeAll(Collections.singleton(null));

                //add the free rawDelivery queries now if we have not gone to GCP for free delivery
                //or we need the nsight free delivery data for comparing with GCP data
                if(!availabilityServiceContextV3.isFreeDeliveryGCPRequestNeeded() || enablePublishFreeDelivery) {
                    for(UUID uuid : fromSiteIdList){
                        deliveryTimeFreeFutureResultSetList.add(getFreeDeliveryEstimatesOptimized(UUID.fromString(shipToSiteID), uuid, splunkLogger));
                    }
                    // if we are doing the free delivery compare, then we need to set the sources used
                    //in the calls to either nsight
                    if(enablePublishFreeDelivery){
                        setSourceLocationsWithInventory(availabilityServiceContextV3);
                    }

                }
            }
        }

        if (availabilityServiceContextV3.getDeliveryEstimateFutures()==null){
            availabilityServiceContextV3.setDeliveryEstimateFutures(new ArrayList<>());
            availabilityServiceContextV3.getDeliveryEstimateFutures().addAll(Futures.inCompletionOrder(deliveryTimeFutureResultSetList));
        }else
            availabilityServiceContextV3.getDeliveryEstimateFutures().addAll(Futures.inCompletionOrder(deliveryTimeFutureResultSetList));

        //add the free rawDelivery futures
        if (availabilityServiceContextV3.getDeliveryFreeEstimateFutures()==null){
            availabilityServiceContextV3.setDeliveryFreeEstimateFutures(new ArrayList<>());
            availabilityServiceContextV3.getDeliveryFreeEstimateFutures().addAll(Futures.inCompletionOrder(deliveryTimeFreeFutureResultSetList));
        }else
            availabilityServiceContextV3.getDeliveryFreeEstimateFutures().addAll(Futures.inCompletionOrder(deliveryTimeFreeFutureResultSetList));
    }

    @Override
    public void setSourceLocationsFromRelatedSites(AvailabilityServiceContextV4 availabilityServiceContextV3){
        Optional<String> sources = null;
        if(availabilityServiceContextV3.getRelatedSites()!=null
                && !availabilityServiceContextV3.getRelatedSites().isEmpty()) {
            Set<String> sourcesSet = new HashSet<>();
            for (RelatedSite relatedSite : availabilityServiceContextV3.getRelatedSites()) {
                SiteDTO siteDTOSource = nSightCache.getSiteIdToSiteMap().get(relatedSite.getSiteId());
                if (siteDTOSource != null && siteDTOSource.getSiteIdentifierList() != null) {
                    Optional<SiteIdentifier> siteIdentifierSource = siteDTOSource.getSiteIdentifierList().stream().filter(siteIdentifier -> siteIdentifier.getName().contains("9 Digit Store Number") || siteIdentifier.getName().contains("DC Abbreviation Code")).findAny();
                    if (siteIdentifierSource.isPresent()) {
                        String source = siteIdentifierSource.get().getValue();
                        sourcesSet.add(source);
                    }
                }
            }
            sources = Optional.of(String.join(",", sourcesSet));
        }
        availabilityServiceContextV3.setSourceLocationsToGCP(sources);
    }

    @Override
    public void setSourceLocationsWithInventory(AvailabilityServiceContextV4 availabilityServiceContextV3){
        Optional<String> sources = null;
        if(availabilityServiceContextV3.getSitesWithInventory()!=null
                && !availabilityServiceContextV3.getSitesWithInventory().isEmpty()) {
            Set<String> sourcesSet = new HashSet<>();
            for (String siteWithInventory : availabilityServiceContextV3.getSitesWithInventory()) {
                SiteDTO siteDTOSource = nSightCache.getSiteIdToSiteMap().get(siteWithInventory);
                if (siteDTOSource != null && siteDTOSource.getSiteIdentifierList() != null) {
                    Optional<SiteIdentifier> siteIdentifierSource = siteDTOSource.getSiteIdentifierList().stream().filter(siteIdentifier -> siteIdentifier.getName().contains("9 Digit Store Number") || siteIdentifier.getName().contains("DC Abbreviation Code")).findAny();
                    if (siteIdentifierSource.isPresent()) {
                        String source = siteIdentifierSource.get().getValue();
                        sourcesSet.add(source);
                    }
                }
            }
            sources = Optional.of(String.join(",", sourcesSet));
        }
        availabilityServiceContextV3.setSourceLocationsToNSight(sources);
    }

    @Override
    public boolean needGCPFreeDeliveryCall(AvailabilityServiceContextV4 availabilityServiceContextV3){
        boolean isGCPRequestNeeded = false;
        if (allFreeCallsToGcp)
            isGCPRequestNeeded = true;
        else if (allFreeCallsToNsight)
            isGCPRequestNeeded =  false;
        else {
            //else we will evaulate
            Optional<SiteIdentifier> locationSiteId = availabilityServiceContextV3.getRequestingSite().getSiteIdentifierList()
                    .stream()
                    .filter(siteIdentifier -> siteIdentifier.getName().compareToIgnoreCase("9 Digit Store Number") == 0 || siteIdentifier.getName().compareToIgnoreCase("DC Abbreviation Code") == 0).findFirst();
            Optional<String> pilotLocationPrefix = Optional.empty();
            Optional<String> pilotLocation = Optional.empty();
            String pilotLocationPrefixAll = availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode() + "-ALL";
            if (locationSiteId.isPresent() && locationSiteId.get().getValue().length() > 2) {
                pilotLocationPrefix = Optional.of(availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode() + "-" + locationSiteId.get().getValue().substring(0, 3));
                pilotLocation = Optional.of(availabilityServiceContextV3.getApplicationID() + "-" + availabilityServiceContextV3.getRequest().getHeaderRequest().getCountryCode() + "-" + locationSiteId.get().getValue());
            }
            isGCPRequestNeeded = (pilotSellingChannelsCallsToGcp && freeDeliveryPilotSellingChannels.contains(availabilityServiceContextV3.getApplicationID())
            )
                    ||
                    (pilotLocationPrefix.isPresent()
                            && (nSightCache.getDeliveryDataSvcPilotStores().toUpperCase().contains(pilotLocationPrefix.get().toUpperCase())
                            || nSightCache.getDeliveryDataSvcPilotStores().toUpperCase().contains(pilotLocationPrefixAll.toUpperCase())
                            || nSightCache.getDeliveryDataSvcPilotStores().toUpperCase().contains(pilotLocation.get().toUpperCase()))
                    );
        }
        //also need to set the sourceLocations which will be used in the calls
        if(!availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllOtherDCs()
                && !availabilityServiceContextV3.getRequest().getDetailRequest().getInventoryRelationshipsRequested().isAllSuppliers()){
            //Now get the sources
            setSourceLocationsFromRelatedSites(availabilityServiceContextV3);
        }
        return isGCPRequestNeeded;
    }

    @Override
    public ListenableFuture<List<DeliveryResponseV1>> setGCPFreeDeliveryFutures(List<DeliveryService> deliveryServiceRequests, SplunkLogger splunkLogger, AvailabilityServiceContextV4 availabilityServiceContextV4){
        //now call the async delivery client
        Stopwatch sw=Stopwatch.createStarted();
        long start = System.currentTimeMillis();
        SettableFuture<List<DeliveryResponseV1>> gcpFreeDeliveryFutures = SettableFuture.create();
        asyncDeliveryClient
                .sendAllRequests(deliveryServiceRequests,splunkLogger)
                .collectList()
                .timeout(Duration.of(deliveryServiceTimeout, MILLIS))
                .onErrorReturn(ImmutableList.of())
                .subscribe(response ->
                {
                    LOGGER.info(splunkLogger.toString() +" createDeliveryCalls: Executed requests in {}, got {} responses",sw.stop(), response.size());
                    gcpFreeDeliveryFutures.set(response);
                });
        LOGGER.info(splunkLogger.toString() + " createDeliveryCalls to GCP sending for all requests in {} ms", System.currentTimeMillis()-start);
        availabilityServiceContextV4.setGcpFreeDeliveryFutures(gcpFreeDeliveryFutures);
        return gcpFreeDeliveryFutures;
    }

    private Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> processPaidDeliveryFutures(Map<String,
            List<com.genpt.nsight.v4.model.RawDelivery>> quickestDeliveryBySite, Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySite,Map<String,
            List<com.genpt.nsight.v4.model.RawDelivery>> paidDeliveryBySite,Set<String> fromSitesToZip,
            AvailabilityServiceContextV4 availabilityServiceContextV3, SplunkLogger splunkLogger) {
        if (availabilityServiceContextV3.getDeliveryEstimateFutures() != null){
            Stopwatch sw = Stopwatch.createStarted();
            Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> deliveryBySite = availabilityServiceContextV3.getDeliveryEstimateFutures().stream().flatMap(resultSetListenableFuture -> {
                ResultSet resultSet = null;
                List<com.genpt.nsight.v4.model.RawDelivery> rawDeliveryList = new ArrayList<>();
                try {
                    resultSet = resultSetListenableFuture.get();
                    while (!resultSet.isExhausted()) {
                        Row row = resultSet.one();
                        String serviceDurationDesignator = row.getString("service_duration_designator");
                        com.genpt.nsight.v4.model.RawDelivery rawDelivery = createDelivery(serviceDurationDesignator, row, availabilityServiceContextV3, rawDeliveryList, fromSitesToZip);
                        if (Boolean.TRUE.equals(availabilityServiceContextV3.isFreePaid())) {
                            createFreePaidDelivery(serviceDurationDesignator, row, quickestDeliveryBySite, availabilityServiceContextV3, freeDeliveryBySite, paidDeliveryBySite);
                        }
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                    LOGGER.error(splunkLogger.toString() + " , Error in setDeliveryBySite, Message :  {}, StackTrace: {} ", ex.getMessage(), ex.getStackTrace());
                }
                return rawDeliveryList.stream();
            }).collect(Collectors.groupingBy(delivery -> delivery.getSiteId()));
            LOGGER.info(splunkLogger.toString() +" Processed NSight Paid Delivery Futures in {} ",sw.stop());
            return deliveryBySite;
        }else
            return null;
    }

    public Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> processFreeDeliveryFutures(AvailabilityServiceContextV4 availabilityServiceContextV3, SplunkLogger splunkLogger,
                                                                                               Set<String> fromSitesToZip){
        if(availabilityServiceContextV3.getDeliveryFreeEstimateFutures()!=null) {
            Stopwatch sw = Stopwatch.createStarted();
            Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySiteMap = availabilityServiceContextV3.getDeliveryFreeEstimateFutures().stream().flatMap(resultSetListenableFuture -> {
                ResultSet resultSet = null;
                List<com.genpt.nsight.v4.model.RawDelivery> freeRawDeliveryList = new ArrayList<>();
                try {
                    resultSet = resultSetListenableFuture.get();
                    while (!resultSet.isExhausted()) {
                        Row row = resultSet.one();
                        String serviceDurationDesignator = row.getString("service_duration_designator");
                        com.genpt.nsight.v4.model.RawDelivery rawDelivery = createFreeDelivery(serviceDurationDesignator, row, availabilityServiceContextV3, freeRawDeliveryList, fromSitesToZip);
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                    LOGGER.error(splunkLogger.toString() + " , Error in setDeliveryBySite, Message :  {}, StackTrace: {} ", ex.getMessage(), ex.getStackTrace());
                }
                return freeRawDeliveryList.stream();
            }).collect(Collectors.groupingBy(delivery -> delivery.getSiteId()));
            LOGGER.info(splunkLogger.toString() +" Processed NSight Free Delivery Futures in {} ",sw.stop());
            //now we can publish these GCP responses to pubsub
            if(enablePublishFreeDelivery){
                Stopwatch swAsync = Stopwatch.createStarted();
                // spawn a thread to publish response asynchronously
                try {
                    //get the destination
                    String destination=null;
                    Optional<SiteIdentifier> siteIdentifierDestination = availabilityServiceContextV3.getRequestingSite().getSiteIdentifierList().stream().filter(siteIdentifier -> siteIdentifier.getName().contains("9 Digit Store Number") || siteIdentifier.getName().contains("DC Abbreviation Code")).findAny();
                    if (siteIdentifierDestination.isPresent()) {
                        destination = siteIdentifierDestination.get().getValue();
                    }
                    asyncDeliveryPubsubPublisher.triggerPublish(freeDeliveryBySiteMap,destination,availabilityServiceContextV3.getSourceLocationsToNSight());
                } catch (Exception e) {
                    LOGGER.error("Exception Publishing Delivery Response from GCP to Pubsub");
                }
                LOGGER.info(splunkLogger.toString()+" Spawned thread to Publish Free Delivery NSight Responses to Pubsub for Compare in {} ",swAsync.stop());
            }
            return freeDeliveryBySiteMap;
        }else
            return null;
    }

    public Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> processFreeDeliveryGCPResponses(AvailabilityServiceContextV4 availabilityServiceContextV3, SplunkLogger splunkLogger){
        if(availabilityServiceContextV3.getGcpFreeDeliveryFutures()!=null) {
            Stopwatch sw = Stopwatch.createStarted();
            Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySiteMap = availabilityServiceContextV3.getGcpFreeDeliveryResponseList().stream().flatMap(deliveryResponseV1 -> {
                List<com.genpt.nsight.v4.model.RawDelivery> freeRawDeliveryList = new ArrayList<>();
                try {
                    deliveryResponseV1.getDeliveryOptions().forEach(deliveryOption -> {
                        String serviceDurationDesignator = deliveryOption.getDurationType();
                        if(serviceDurationDesignator!=null) {
                            serviceDurationDesignator = WordUtils.capitalize(serviceDurationDesignator.toLowerCase());
                            createFreeDeliveryGCP(serviceDurationDesignator, deliveryOption, availabilityServiceContextV3, freeRawDeliveryList);
                        }
                    });
                } catch (Exception ex) {
                    ex.printStackTrace();
                    LOGGER.error(splunkLogger.toString() + " , Error in setDeliveryBySite, Message :  {}, StackTrace: {} ", ex.getMessage(), ex.getStackTrace());
                }
                return freeRawDeliveryList.stream();
            }).collect(Collectors.groupingBy(delivery -> delivery.getSiteId()));
            LOGGER.info(splunkLogger.toString() +" Transformed GCP Futures to Free Delivery Response in {} ",sw.stop());
            return freeDeliveryBySiteMap;
        }else
            return null;
    }

    private Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> processFreeDeliveryGCPFutures(AvailabilityServiceContextV4 availabilityServiceContextV4){
        com.genpt.nsight.v2.SplunkLogger splunkLogger = new com.genpt.nsight.v2.SplunkLogger(availabilityServiceContextV4.getRequest().getHeaderRequest());
        Stopwatch sw = Stopwatch.createStarted();
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> rawDeliveryMap = null;
        if(availabilityServiceContextV4.getGcpFreeDeliveryFutures() !=null) {
            List<DeliveryResponseV1> deliveryResponseList = processAllFreeGCPResponses(availabilityServiceContextV4.getGcpFreeDeliveryFutures());
            availabilityServiceContextV4.setGcpFreeDeliveryResponseList(deliveryResponseList);
            //now start processing those
            rawDeliveryMap = processFreeDeliveryGCPResponses(availabilityServiceContextV4,splunkLogger);
        }
        LOGGER.info(splunkLogger.toString()+" Finished processing All Free Delivery GCP Async Futures in {} ",sw.stop());
        return rawDeliveryMap;
    }

    private List<DeliveryResponseV1> processAllFreeGCPResponses(ListenableFuture<List<DeliveryResponseV1>> deliveryFlux) {
        return Futures.getUnchecked(deliveryFlux);
        //TODO error handling
    }

    @Override
    public void setDeliveryBySite(AvailabilityServiceContextV4 availabilityServiceContextV3) {
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());

        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> quickestDeliveryBySite = new HashMap<>();
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySite = new HashMap<>();
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> paidDeliveryBySite = new HashMap<>();
        Map<String, Map<Integer, List<com.genpt.nsight.v4.model.Delivery>>> deliveryBySiteByDay = new HashMap<>();

/*        if (Boolean.TRUE.equals(availabilityServiceContextV3.getRequest().getActionRequest().getAction().getNeedQuickestDelivery())){
            setOneDeliveryBySite(availabilityServiceContextV3);
        }else */
        {
            //create a set where we store all the 5 Digit deliveries
            Set<String> fromSitesToZip = new HashSet<>();
            //iterate over the futures from site to site rawDelivery paid table
            Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> deliveryBySite = processPaidDeliveryFutures(quickestDeliveryBySite,freeDeliveryBySite,paidDeliveryBySite,fromSitesToZip,availabilityServiceContextV3,splunkLogger);
            availabilityServiceContextV3.setDeliveryBySite(deliveryBySite);

            // we might need to iterate over the futures from the new free rawDelivery table too only if its not for zip codes
            Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySiteMap = null;
            if(!availabilityServiceContextV3.isFreeDeliveryGCPRequestNeeded() || enablePublishFreeDelivery) {
                if(!enablePublishFreeDelivery)// we need the nsight free deliveries in the response
                    freeDeliveryBySiteMap = processFreeDeliveryFutures(availabilityServiceContextV3, splunkLogger, fromSitesToZip);
                else if(!availabilityServiceContextV3.isFreeDeliveryGCPRequestNeeded())
                    freeDeliveryBySiteMap = processFreeDeliveryFutures(availabilityServiceContextV3, splunkLogger, fromSitesToZip);
                else//we just need to process the nsight free deliveries for the compare
                    processFreeDeliveryFutures(availabilityServiceContextV3, splunkLogger, fromSitesToZip);
            }
            // check if we had requested GCP deliveries and use those in the response
            if(availabilityServiceContextV3.isFreeDeliveryGCPRequestNeeded())
                freeDeliveryBySiteMap = processFreeDeliveryGCPFutures(availabilityServiceContextV3);

            //we need to merge the two Maps. iterate over the paid map and add the free map and merge them
            for (Map.Entry<String, List<com.genpt.nsight.v4.model.RawDelivery>> entry: freeDeliveryBySiteMap.entrySet()){
                String keySiteID = entry.getKey();
                List<com.genpt.nsight.v4.model.RawDelivery> freeRawDeliveryList = entry.getValue();
                List<com.genpt.nsight.v4.model.RawDelivery> paidRawDeliveryList = deliveryBySite.get(keySiteID);

                if (paidRawDeliveryList != null && paidRawDeliveryList.size() >0){// does this site alread have paid
                    paidRawDeliveryList.addAll(freeRawDeliveryList);
                }else
                    deliveryBySite.put(keySiteID, freeRawDeliveryList);
            }

            availabilityServiceContextV3.setDeliveryBySite(deliveryBySite);

            if (Boolean.TRUE.equals(availabilityServiceContextV3.isFreePaid())) {
                availabilityServiceContextV3.setQuickestFreePaidDeliveryBySite(quickestDeliveryBySite);
                availabilityServiceContextV3.setFreeDeliveryBySite(freeDeliveryBySite);
                availabilityServiceContextV3.setPaidDeliveryBySite(paidDeliveryBySite);
            }

            //populate by site by day into a new map. This is for V4 where we store based on deliver day
            //this should be the Zone of calling store
            ZonedDateTime now = null;
            if (availabilityServiceContextV3.getStoreTimeZone().contains(".")) {
                now = availabilityServiceContextV3.getStoreCurrentTime().withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset());
            }
            else
                now = availabilityServiceContextV3.getStoreCurrentTime().withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset());

            ZonedDateTime beginningOfDay = now.truncatedTo(ChronoUnit.DAYS);
            ZonedDateTime nowUTC = ZonedDateTime.now(ZoneId.of("UTC"));
            final ZonedDateTime beginningOfDayUTC = nowUTC.truncatedTo(ChronoUnit.DAYS);

            // these are paid  and free deliveries that we are iterating. this is for a site the list by type
            for (Map.Entry<String, List<com.genpt.nsight.v4.model.RawDelivery>> entry : deliveryBySite.entrySet()) {
                    //get from the raw data
                    List<com.genpt.nsight.v4.model.RawDelivery> deliveriesByType = entry.getValue();
                    String siteID = entry.getKey();
                    //for this site get the List by Day
                    Map<Integer, List<com.genpt.nsight.v4.model.Delivery>> deliveryByDayMap= deliveryBySiteByDay.get(siteID);
                    //for the day if we dont have one, create it
                    if (deliveryByDayMap == null)
                        deliveryByDayMap =  new HashMap<>();

                    //from the raw data iterate over the deliveries by type for this site
                    for (com.genpt.nsight.v4.model.RawDelivery rawDelivery : deliveriesByType) {
                        //iterate over each cutoff time from the raw data
                        if (rawDelivery.getDeliveryCutOff() != null) {
                            for (com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff : rawDelivery.getDeliveryCutOff()) {
                                boolean afterCutOff = false;
                                String deliveryDayOfWeek = DeliveryCutOffSerializer.DAY_OF_WEEK[ deliveryCutOff.getDayOfTheWeek()];
                                //get the timezone to check if its after midnight which means its really late for that day
                                float shipFromTZ = 0;
                                boolean isNextDay =  false;
                                boolean isEarliestNextDay = false;
                                boolean localCrossedUTCMidnight = false;

                                //Since Store console uses the ship from site timezone  to calculate the UTC cutoff's we will use the same for conversions to local time
                                //String tz = nSightCache.getSiteIdToSiteMap().get(siteID).getTimeZone();
                                String tz = nSightCache.getSiteIdToSiteMap().get(availabilityServiceContextV3.getRequestingSite().getSiteIdentifier()).getTimeZone();

                                if (tz != null && tz.trim()!= "" && tz.trim().compareToIgnoreCase("NaN")!= 0 && nSightCache.getSiteIdToSiteMap().get(siteID)!=null && nSightCache.getSiteIdToSiteMap().get(siteID).getTimeZone()!= null && nSightCache.getSiteIdToSiteMap().get(siteID).getTimeZone().trim()!= "")
                                	shipFromTZ = Float.parseFloat(nSightCache.getSiteIdToSiteMap().get(siteID).getTimeZone());
                                if (availabilityServiceContextV3.getStoreOffSetTime()!=null && shipFromTZ !=0) {
                                    isNextDay = availabilityServiceContextV3.isCutOffNextDay(deliveryCutOff.getCutOffTime(), shipFromTZ);
                                    localCrossedUTCMidnight = availabilityServiceContextV3.isStorePastMidnightUTC();
                                }

                                // if processing a current day cutoff, check if its past the cutoff time
                                if (deliveryCutOff.getDayOfTheWeek() == availabilityServiceContextV3.getDayOfWeekLocal()) {
                                	
                                	//getting cutoff time with date so that we can compare the time with date also in order to handle different time zone issue
                                 	String cutOffDateTime = null;
                                	if (localCrossedUTCMidnight && !isNextDay) {
                                        ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                        cutOffDateTime = availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getCutOffTime(), beginningOfDayBack, isNextDay, "Days", deliveryDayOfWeek, localCrossedUTCMidnight);
                                    }else
                                    	cutOffDateTime = availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getCutOffTime(), beginningOfDayUTC, isNextDay, "Days", deliveryDayOfWeek, localCrossedUTCMidnight);

                                	
                                    //we are processing a current day cutoff
                                	if (availabilityServiceContextV3.getStoreOffSetTime() != null && shipFromTZ != 0 && nSightCache.getSiteIdToSiteMap().get(siteID)!=null && nSightCache.getSiteIdToSiteMap().get(siteID).getTimeZoneID()!=null) {// this needs to work for days, minutes, hours, etc., minutes, hours, etc.
//                                        if (!availabilityServiceContextV3.requestTimeBeforeCutOffTime(deliveryCutOff.getCutOffTime(), shipFromTZ, deliveryCutOff.getDeliveryEstimate().getDurationMinValue()))//expects millis
                                		if (!availabilityServiceContextV3.requestDateTimeBeforeCutOffTime(cutOffDateTime, shipFromTZ, deliveryCutOff.getDeliveryEstimate().getDurationMinValueWithLeadTime(),nSightCache.getSiteIdToSiteMap().get(siteID)))//need to use lead times
                                            afterCutOff = true;
                                    } else {
                                        long nowCurrentDay = 1000000 * availabilityServiceContextV3.getTimeOfDay();
                                        if (deliveryCutOff.getCutOffTime() < nowCurrentDay)
                                            afterCutOff = true;
                                    }
                                }
                                if (afterCutOff)
                                    continue;//not interested
                                //Now create a new V4 RawDelivery Object
                                com.genpt.nsight.v4.model.Delivery deliveryDay = new com.genpt.nsight.v4.model.Delivery();//
                                deliveryDay.setDeliveryType(rawDelivery.getDeliveryType());
                                deliveryDay.setZone(rawDelivery.getZone());

                                //this is needed only internally
                                deliveryDay.setDeliveryDestination(rawDelivery.getDeliveryDestination());

                                if (rawDelivery.getDeliveryCategory()!=null && rawDelivery.getDeliveryCategory().length() >0)
                                    deliveryDay.setDeliveryCategory(rawDelivery.getDeliveryCategory());
                                else
                                    deliveryDay.setDeliveryCategory("Ranged");

                                deliveryDay.setCutOffTime(deliveryCutOff.getCutOffTime());
                                deliveryDay.setDayOfWeek(deliveryCutOff.getDayOfTheWeek());

                                if (localCrossedUTCMidnight && !isNextDay) {
                                    ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                    deliveryDay.setCutOffDateTime(availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getCutOffTime(), beginningOfDayBack, isNextDay, "Days", deliveryDayOfWeek, localCrossedUTCMidnight));
                                }else
                                    deliveryDay.setCutOffDateTime(availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getCutOffTime(), beginningOfDayUTC, isNextDay, "Days", deliveryDayOfWeek, localCrossedUTCMidnight));

                                try {
                                    deliveryDay.setCutOffTimeString(nSightCache.getDeliveryCutOffTimeStore().get(deliveryCutOff.getCutOffTime()));
                                } catch (ExecutionException e) {
                                    e.printStackTrace();
                                }
                                deliveryDay.setDayofweekString(DeliveryCutOffSerializer.DAY_OF_WEEK[deliveryDay.getDayOfWeek()]);

                                if (nSightProperties.getPaidMOTS().contains(rawDelivery.getDeliveryType()))
                                    deliveryDay.setFreeDelivery(false);
                                else
                                    deliveryDay.setFreeDelivery(true);

                                if (rawDelivery.getDeliveryEstimate()!=null) {
                                    DeliveryEstimateFreePaid deliveryEstimateFreePaid = new DeliveryEstimateFreePaid();
                                    deliveryEstimateFreePaid.setDurationDesignator(deliveryCutOff.getDeliveryEstimate().getDurationDesignator());

                                    //if we use lead times
                                    deliveryEstimateFreePaid.setDurationMaxValue(deliveryCutOff.getDeliveryEstimate().getDurationMaxValueWithLeadTime());
                                    deliveryEstimateFreePaid.setDurationMinValue(deliveryCutOff.getDeliveryEstimate().getDurationMinValueWithLeadTime());

                                    //add the calculated fields
                                    //check if this is a On Demand category
                                    if (deliveryDay.getDeliveryCategory().compareToIgnoreCase("On Demand")==0 && deliveryCutOff.getEarliesTime()!=-1) {
                                    /// check if the current time is before earliest time. then use earlies time
                                        String earliestTimeAsString = null;
                                        ZonedDateTime nowEarliest = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId()));
                                        //checking if earliest time has crossed midnight and calculate the earliest time
                                        isEarliestNextDay = availabilityServiceContextV3.isEarliestNextDay(deliveryCutOff.getEarliesTime(), shipFromTZ);
                                        if (isEarliestNextDay) {
                                            earliestTimeAsString = availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getEarliesTime(), beginningOfDayUTC, true, "Days", deliveryDayOfWeek,localCrossedUTCMidnight);
                                        }
                                        else {
                                            if (localCrossedUTCMidnight  && !isNextDay) {
                                                ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                                earliestTimeAsString = availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getEarliesTime(), beginningOfDayBack, false, "Days", deliveryDayOfWeek, localCrossedUTCMidnight);
                                            }
                                            else
                                                earliestTimeAsString = availabilityServiceContextV3.getDateTime(0, deliveryCutOff.getEarliesTime(), beginningOfDayUTC, false, "Days", deliveryDayOfWeek,localCrossedUTCMidnight);
                                        }
                                        ZonedDateTime earliestTime = availabilityServiceContextV3.parseDateTime(earliestTimeAsString, false);

                                        // now calculate the expectation based on either earliest time or current time
                                        if (availabilityServiceContextV3.getStoreCurrentTime().withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset()).withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset()).isBefore(earliestTime)){//current time before earliest time then to ET+offsets
                                            if (localCrossedUTCMidnight  && !isEarliestNextDay) {
                                                ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                                deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), deliveryCutOff.getEarliesTime(), beginningOfDayBack, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                                deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), deliveryCutOff.getEarliesTime(), beginningOfDayBack, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            }else{
                                                deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), deliveryCutOff.getEarliesTime(), beginningOfDayUTC, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                                deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), deliveryCutOff.getEarliesTime(), beginningOfDayUTC, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            }
                                        }else {//current time after ET then set to CurrentTime + offset
                                            if (localCrossedUTCMidnight  && !isEarliestNextDay) {
                                                ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                                deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), availabilityServiceContextV3.getMilliSecondsFromMidnight(now), beginningOfDayBack, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                                deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), availabilityServiceContextV3.getMilliSecondsFromMidnight(now), beginningOfDayBack, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            }else{
                                                deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), availabilityServiceContextV3.getMilliSecondsFromMidnight(now), beginningOfDay, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                                deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), availabilityServiceContextV3.getMilliSecondsFromMidnight(now), beginningOfDay, isEarliestNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            }
                                        }
                                    }else {// it is scheduled , Ranged
                                        if (localCrossedUTCMidnight  && !isNextDay) {
                                            ZonedDateTime beginningOfDayBack = beginningOfDayUTC.minusDays(1);
                                            deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), deliveryCutOff.getCutOffTime(), beginningOfDayBack, isNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), deliveryCutOff.getCutOffTime(), beginningOfDayBack, isNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                        } else {
                                            deliveryEstimateFreePaid.setExpectationRangeMin(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMinValue(), deliveryCutOff.getCutOffTime(), beginningOfDayUTC, isNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                            deliveryEstimateFreePaid.setExpectationRangeMax(availabilityServiceContextV3.getDateTime(deliveryEstimateFreePaid.getDurationMaxValue(), deliveryCutOff.getCutOffTime(), beginningOfDayUTC, isNextDay, deliveryEstimateFreePaid.getDurationDesignator(), deliveryDayOfWeek, localCrossedUTCMidnight));
                                        }
                                    }
                                    //factor in store hours with the expectations

                                    availabilityServiceContextV3.translateTimesToLocal( deliveryDay,  siteID, deliveryEstimateFreePaid, false);
                                    //////
                                    if(!afterCutOff) {
                                        //check the expectations and see if there are free delivereies  that fall when the store is closed
                                        //then move them to the next store open + offset. if in days then move to the next store close(free)
                                        CalculateExpectationsBasedOnStoreHours( deliveryEstimateFreePaid,  availabilityServiceContextV3,  deliveryDay, siteID);

                                        deliveryDay.setDeliveryEstimate(deliveryEstimateFreePaid);
                                    }
                                }
                                //check if a entry for the list exists for this by day
                                List<com.genpt.nsight.v4.model.Delivery> listByDay = deliveryByDayMap.get(deliveryCutOff.getDayOfTheWeek());
                                if(!afterCutOff) {
                                    if (listByDay == null)
                                        listByDay = new ArrayList<>();
                                    listByDay.add(deliveryDay);// add to day list
                                    deliveryByDayMap.put(deliveryCutOff.getDayOfTheWeek(), listByDay);
                                }
                            }//iterate deliveries cutoffs by day
                        }
                    }//iterate by rawDelivery type
                    //done with site. sort it for this site now
                    List<com.genpt.nsight.v4.model.Delivery> currentSitesDeliveries  = new ArrayList<>();
                    List<com.genpt.nsight.v4.model.Delivery> currentSitesDeliveriesLimited  = new ArrayList<>();

                    //add all deliveries for ALL days of the week
                    for (List<com.genpt.nsight.v4.model.Delivery> value : deliveryByDayMap.values())
                        currentSitesDeliveries.addAll(value);
                    //sort them by the arrival datetime
                    Collections.sort(currentSitesDeliveries);

                    //now put them in the map back after populating the arrival sequence number
                    //TreeMap<Integer, List<com.genpt.nsight.v4.model.RawDelivery>> deliveryByDayMapSorted= new TreeMap<>();
                    //iterate again over the options to update properties
                    int i =0;
                    boolean quickestPaid = false;
                    boolean quickestFree = false;
                    Map<String, Integer> deliveryCountMap = new HashMap<>();
                    //iterate ove rthe sorteed list of deliveries for this site
                    for ( com.genpt.nsight.v4.model.Delivery currentSiteDelivery : currentSitesDeliveries){
/*                        List<com.genpt.nsight.v4.model.RawDelivery> listByDay = deliveryByDayMapSorted.get(currentSiteDelivery.getDayOfWeek());
                        if (listByDay == null)
                            listByDay = new ArrayList<>();*/
                        int total = 0;
                        Integer integer = deliveryCountMap.get(currentSiteDelivery.getDeliveryType());
                        if (integer!=null)
                            total=integer;
                        if (total > 1)//since we are allowing a max of 2 per delivery type
                            continue;

                        deliveryCountMap.put(currentSiteDelivery.getDeliveryType(),++total);

                        currentSiteDelivery.setDeliveryArrivalSeq(i+1);
                        //set the quickest options
                        if (i==0){// the first delivery in the list
                            currentSiteDelivery.setQuickestOption(true);
                            if (!currentSiteDelivery.isFreeDelivery())//not a free delivery
                                quickestPaid = true;
                            else
                                quickestFree = true;
                        }else{//subsequent deliveries
                            if (quickestFree == false || quickestPaid == false) {//still havent figured the quickest for both
                                if (quickestPaid == false && !currentSiteDelivery.isFreeDelivery()){//its a paid delivery
                                    currentSiteDelivery.setQuickestOption(true);
                                    quickestPaid = true;//found the quickest paid
                                }
                                if (quickestFree == false && currentSiteDelivery.isFreeDelivery()) {
                                    currentSiteDelivery.setQuickestOption(true);
                                    quickestFree = true;
                                }
                            }
                        }
                        // need to add lead times
                        List<LeadTimes> leadTimesforSite = availabilityServiceContextV3.getLeadTimeBySite().get(siteID);
                        if (leadTimesforSite != null && leadTimesforSite.size() > 0) {
                            ArrayList<com.genpt.nsight.v4.model.LeadTimes> leadTimesList = new ArrayList<>();
                            for (  LeadTimes leadTimesRaw : leadTimesforSite) {// we need for the day of the week and service type
                                if (leadTimesRaw.getDayOfTheWeek().compareToIgnoreCase(currentSiteDelivery.getDayofweekString()) == 0 && leadTimesRaw.getDeliveryServiceType().compareToIgnoreCase(currentSiteDelivery.getDeliveryType())==0) {
                                    com.genpt.nsight.v4.model.LeadTimes leadTimesNew = new com.genpt.nsight.v4.model.LeadTimes();
                                    leadTimesNew.setLeadTime(leadTimesRaw.getLeadTime());
                                    leadTimesNew.setReason(leadTimesRaw.getReason());
                                    leadTimesNew.setDurationDesignator(leadTimesRaw.getDurationDesignator());
                                    leadTimesList.add(leadTimesNew);
                                }
                            }
                            if (leadTimesList.size() >0)//add only if there are values
                                currentSiteDelivery.setLeadTimesList(leadTimesList);
                        }
                        currentSitesDeliveriesLimited.add(currentSiteDelivery);
                        i++;
                    }
                    //At this point we have the limited list of deliveries for this sorted in ascending order
                    availabilityServiceContextV3.getDeliveryBySiteSorted().put(siteID, currentSitesDeliveriesLimited);
            }//done with all sites
        }
    }

    public void CalculateExpectationsBasedOnStoreHours(DeliveryEstimateFreePaid deliveryEstimateFreePaid, AvailabilityServiceContextV4 availabilityServiceContextV3, Delivery deliveryDay, String SiteID){
        boolean updatedExpectations = false;
        ArrayList< StoreHours> storeHoursList= availabilityServiceContextV3.getRequestingSiteStoreHoursList();//map for each day of the week
        if (deliveryDay.isFreeDelivery() && deliveryEstimateFreePaid.getExpectations() != null && deliveryEstimateFreePaid.getExpectations().size() > 0) {//have delivery data and only free
            for ( Expectation expectation : deliveryEstimateFreePaid.getExpectations()) {//go over all expectations
                if (expectation.getExpectationRangeMin() != null) {
                    ZonedDateTime zonedDateTime = availabilityServiceContextV3.parseDateTime(expectation.getExpectationRangeMin(), false);
                    int index = zonedDateTime.getDayOfWeek().getValue()-1;//day of week is 1 based starting with Monday
                    String expectationDay = DeliveryCutOffSerializer.DAY_OF_WEEK[index];
                    //int foundIndex = availabilityServiceContextV3.getDaysAfterToday(expectationDay);//need to keep a count of how many days after today the expectation might be moved to
                    long foundIndex = ChronoUnit.DAYS.between(availabilityServiceContextV3.getStoreCurrentTime().toLocalDate(), zonedDateTime.toLocalDate());
                    StoreHours storeHours = storeHoursList.get(index);//start looking from that delivery day

                    if (storeHours == null) {//not open on that day, start looking from next day
                        ++foundIndex; //since we are starting to look from the next day
                        index  = (index == 6) ? index=0 :  ++index;//store hours is zero based list starting from Monday
                        for (int i = 0; i < 6; i++) {
                            StoreHours storeHoursLoop = storeHoursList.get(index);//start looking from next day
                            if (storeHoursLoop!=null){//is open
                                // use the store open hours if its a delivery that is in anything other than days
                                if (deliveryEstimateFreePaid.getDurationDesignator().compareToIgnoreCase("Days")==0 ){
                                    ZonedDateTime nextStoreClose = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withHour(storeHoursLoop.getClose().getHour()).withMinute(storeHoursLoop.getClose().getMinute()).withSecond(0).minusMinutes(availabilityServiceContextV3.getStoreClosingOffset()).plusDays(foundIndex);
                                    String nextStoreCloseString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreClose, 0, availabilityServiceContextV3.isUseUTC());
                                    expectation.setExpectationRangeMin(nextStoreCloseString);
                                    expectation.setExpectationRangeMax(nextStoreCloseString);
                                }else{//minutes or hours
                                    if (expectation.getOffsetValue()!=null) {//we have an offset
                                        ZonedDateTime nextStoreOpen = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withHour(storeHoursLoop.getOpen().getHour()).withMinute(storeHoursLoop.getOpen().getMinute()).withSecond(0).plusDays(foundIndex);
                                        String nextStoreOpenString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreOpen, Integer.parseInt(expectation.getOffsetValue()), availabilityServiceContextV3.isUseUTC());
                                        expectation.setExpectationRangeMin(nextStoreOpenString);
                                        expectation.setExpectationRangeMax(nextStoreOpenString);
                                    }else {
                                        ZonedDateTime nextStoreOpen = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withHour(storeHoursLoop.getOpen().getHour()).withMinute(storeHoursLoop.getOpen().getMinute()).withSecond(0).plusDays(foundIndex);
                                        String nextStoreOpenString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreOpen, 0, availabilityServiceContextV3.isUseUTC());
                                        expectation.setExpectationRangeMin(nextStoreOpenString);
                                        expectation.setExpectationRangeMax(nextStoreOpenString);
                                    }
                                }
                                updatedExpectations=true;//expectations were updated
                                break;
                            }else{
                                //need to get the next store open from that day
                                index  = (index == 6) ? index=0 :  ++index;
                                foundIndex++;
                            }
                        }
                    }else{//open on that day but need to check if expectation is within calculate store hours for that day
                        // Need to do this for all non today deliveries since today deliveries would have been taken care off in populateexpectations
                        ZonedDateTime originalMinExpectation = availabilityServiceContextV3.parseDateTime(deliveryEstimateFreePaid.getExpectationRangeMin(), false);
                        boolean isTodayDelivery =  originalMinExpectation.toLocalDate().equals(LocalDate.now(originalMinExpectation.getZone()));//is it todays delivery
                        if (!isTodayDelivery) {
                            //ZonedDateTime quickest = deliveryEstimateFreePaid.getQuickestExpectationOfAllMin();//of all the expectations get the quickest one
                            ZonedDateTime quickest = availabilityServiceContextV3.parseDateTime(expectation.getExpectationRangeMin(), false);//of all the expectations get the quickest one
                            //get the calculated store close for that day
                            ZonedDateTime StoreCloseThatDay = ZonedDateTime.now(TimeZone.getTimeZone(availabilityServiceContextV3.getStoreZoneID()).toZoneId()).withYear(zonedDateTime.getYear()).withMonth(zonedDateTime.getMonthValue()).withDayOfMonth(zonedDateTime.getDayOfMonth()).withHour(storeHours.getClose().getHour()).withMinute(storeHours.getClose().getMinute()).withSecond(0).minusMinutes(availabilityServiceContextV3.getStoreClosingOffset());
                            foundIndex=0;
                            // we cannot directly compare the 2 datetimes since the expectation is based of the offset from site table which never changes and the store close is based of ZoneID from Timezone table
                            if (quickest.withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset()).isAfter(StoreCloseThatDay.withZoneSameLocal(availabilityServiceContextV3.getStoreZoneOffset()))) {// move to next store open day
                                ++foundIndex; //since we are starting to look from the next day
                                index  = (index == 6) ? index=0 :  ++index;
                                for (int i = 0; i < 6; i++) {
                                    StoreHours storeHoursLoop = storeHoursList.get(index);//start looking from next day
                                    if (storeHoursLoop!=null){//is open
                                        // use the store open hours if its a delivery that is in anything other than days
                                        if (deliveryEstimateFreePaid.getDurationDesignator().compareToIgnoreCase("Days")==0 ){
                                            ZonedDateTime nextStoreClose = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withYear(zonedDateTime.getYear()).withMonth(zonedDateTime.getMonthValue()).withDayOfMonth(zonedDateTime.getDayOfMonth()).withHour(storeHoursLoop.getClose().getHour()).withMinute(storeHoursLoop.getClose().getMinute()).withSecond(0).minusMinutes(availabilityServiceContextV3.getStoreClosingOffset()).plusDays(foundIndex);
                                            String nextStoreCloseString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreClose, 0, availabilityServiceContextV3.isUseUTC());
                                            expectation.setExpectationRangeMin(nextStoreCloseString);
                                            expectation.setExpectationRangeMax(nextStoreCloseString);
                                        }else{//minutes or hours
                                            if (expectation.getOffsetValue()!=null) {//we have an offset
                                                ZonedDateTime nextStoreOpen = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withYear(zonedDateTime.getYear()).withMonth(zonedDateTime.getMonthValue()).withDayOfMonth(zonedDateTime.getDayOfMonth()).withHour(storeHoursLoop.getOpen().getHour()).withMinute(storeHoursLoop.getOpen().getMinute()).withSecond(0).plusDays(foundIndex);
                                                String nextStoreOpenString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreOpen, Integer.parseInt(expectation.getOffsetValue()), availabilityServiceContextV3.isUseUTC());
                                                expectation.setExpectationRangeMin(nextStoreOpenString);
                                                expectation.setExpectationRangeMax(nextStoreOpenString);
                                            }else {
                                                ZonedDateTime nextStoreOpen = ZonedDateTime.now(ZoneId.of(availabilityServiceContextV3.getStoreZoneOffset().getId())).withYear(zonedDateTime.getYear()).withMonth(zonedDateTime.getMonthValue()).withDayOfMonth(zonedDateTime.getDayOfMonth()).withHour(storeHoursLoop.getOpen().getHour()).withMinute(storeHoursLoop.getOpen().getMinute()).withSecond(0).plusDays(foundIndex);
                                                String nextStoreOpenString = availabilityServiceContextV3.calculateUTCorLocalTimeBasedOnOffset(nextStoreOpen, 0, availabilityServiceContextV3.isUseUTC());
                                                expectation.setExpectationRangeMin(nextStoreOpenString);
                                                expectation.setExpectationRangeMax(nextStoreOpenString);
                                            }
                                        }
                                        updatedExpectations=true;//expectations were updated
                                        break;
                                    }else{
                                        //need to get the next store open from that day
                                        index  = (index == 6) ? index=0 :  ++index;
                                        foundIndex++;
                                    }
                                }//for
                            }// quickest was after store hours
                        }// other than today delivery
                    }
                }
            }
            /// recalculate the quickest expectations
            if (updatedExpectations) {
                deliveryEstimateFreePaid.setQuickestExpectationOfAllMin(null);// reset it so it gets updated
                availabilityServiceContextV3.assignQuickestExpectation(deliveryEstimateFreePaid, deliveryEstimateFreePaid.getExpectations(), true);
            }
        }
    }

    public void setFreePaidDeliveryBySite(AvailabilityServiceContextV4 availabilityServiceContextV3) {
        //iterate over all the deliveries to the requesting site
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> quickestDeliveryBySite = new HashMap<>();
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySite = new HashMap<>();
        Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> paidDeliveryBySite = new HashMap<>();

        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());

        for (ListenableFuture<ResultSet> resultSetListenableFuture : availabilityServiceContextV3.getDeliveryEstimateFutures()) {
            ResultSet resultSet = null;
            String fromSiteID = null;

            try {
                resultSet = resultSetListenableFuture.get();
                //iterate results
                while (!resultSet.isExhausted()) {
                    Row row = resultSet.one();
                    String serviceDurationDesignator = row.getString("service_duration_designator");
                    createFreePaidDelivery(serviceDurationDesignator,row,quickestDeliveryBySite, availabilityServiceContextV3,freeDeliveryBySite,paidDeliveryBySite);
                }
            } catch (Exception ex) {
                ex.printStackTrace();
                LOGGER.error(splunkLogger.toString() + " , Error in setOneDeliveryBySite, Message :  {}, StackTrace: {} ", ex.getMessage(), ex.getStackTrace());
            }

        }
        availabilityServiceContextV3.setQuickestFreePaidDeliveryBySite(quickestDeliveryBySite);
        availabilityServiceContextV3.setFreeDeliveryBySite(freeDeliveryBySite);
        availabilityServiceContextV3.setPaidDeliveryBySite(paidDeliveryBySite);
    }

    private boolean isDeliveryQuicker( com.genpt.nsight.v4.model.RawDelivery rawDelivery,  com.genpt.nsight.v4.model.RawDelivery existingRawDelivery){
        boolean quicker = false;
        String existingDesignator = existingRawDelivery.getDeliveryEstimate().getDurationDesignator();
        String currentDesignator = existingRawDelivery.getDeliveryEstimate().getDurationDesignator();
        long existCommonDenom = 0;long currentCommonDenom = 0;
        if (existingDesignator.compareToIgnoreCase("Days")==0) {//convert to minutes
            //existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValue() * 1440;
            //use lead times
            existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() * 1440;

        }
        else if (existingDesignator.compareToIgnoreCase("Hours")==0)// hours to inutes
            existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValue() * 60;
        else
            existCommonDenom =  existingRawDelivery.getDeliveryEstimate().getDurationMinValue();

        if (currentDesignator.compareToIgnoreCase("Days")==0) {//convert to minutes
            //currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValue() * 1440;
            //use lead times
            currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() * 1440;
        }
        else if (currentDesignator.compareToIgnoreCase("Hours")==0)// hours to inutes
            currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValue() * 60;
        else
            currentCommonDenom =  rawDelivery.getDeliveryEstimate().getDurationMinValue();
        if (currentCommonDenom < existCommonDenom)
            quicker = true;
        return quicker;
    }

    private void compareAndSetQuickestDeliveryList(AvailabilityServiceContextV4 availabilityServiceContextV3, Row row, List<com.genpt.nsight.v4.model.RawDelivery> existingRawDeliveryList, String deliveryServiceType, String fromSiteID, String serviceDurationDesignator,
                                                     float shipFromTZ, boolean isPaidDelivery, List<com.genpt.nsight.v4.model.RawDelivery> existingFreeRawDeliveryList, List<com.genpt.nsight.v4.model.RawDelivery> existingPaidRawDeliveryList){

        com.genpt.nsight.v4.model.RawDelivery existingRawDelivery = existingRawDeliveryList.get(0);
        com.genpt.nsight.v4.model.RawDelivery existingFreeRawDelivery = existingFreeRawDeliveryList.size() > 0 ? existingFreeRawDeliveryList.get(0) : null;
        com.genpt.nsight.v4.model.RawDelivery existingPaidRawDelivery = existingPaidRawDeliveryList.size() > 0 ? existingPaidRawDeliveryList.get(0): null;

        //create the new rawDelivery
        com.genpt.nsight.v4.model.RawDelivery rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
        rawDelivery.setDeliveryType(deliveryServiceType);
        rawDelivery.setSiteId(fromSiteID);

        List<com.genpt.nsight.v4.model.DeliveryCutOff> deliveryCutOffList = null;
        for (int i = 0; i < 7; i++) {
            //check if the time is a null value
            boolean isNull = row.isNull(dayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7]);
            if (!isNull) {
                long deliveryCutOffTime = row.getTime(dayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7]);
                //include lead times in calculation
                String dayOfWeek = leadTimesDayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7];
                int sumLeadTimes = availabilityServiceContextV3.getLeadTimesForSite(dayOfWeek, fromSiteID, deliveryServiceType);

                boolean afterCutOff = false;
                if (availabilityServiceContextV3.getStoreOffSetTime() != null && shipFromTZ != 0 && nSightCache.getSiteIdToSiteMap().get(fromSiteID)!=null && nSightCache.getSiteIdToSiteMap().get(fromSiteID).getTimeZoneID()!=null) {
                	if (!availabilityServiceContextV3.requestTimeBeforeCutOffTime(deliveryCutOffTime/1000000, shipFromTZ,row.getInt("service_duration_min_value")+sumLeadTimes,nSightCache.getSiteIdToSiteMap().get(fromSiteID)))//expects millis
                        afterCutOff=true;
                }else{
                    long now = 1000000 * availabilityServiceContextV3.getTimeOfDay();
                    if (deliveryCutOffTime < now)
                        afterCutOff=true;
                }

                //if we have passed the cutoff time
                if (i == 0 ) {
                    if (afterCutOff)
                        continue;
                    else{//before the cutoff time for today
                        com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
                        deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff((availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7, deliveryCutOffTime / 1000000L);
                        if (deliveryCutOffList == null) {
                            deliveryCutOffList = new ArrayList<>();
                        }
                        deliveryCutOffList.add(deliveryCutOff);
                        DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
                        rawDelivery.setDeliveryEstimate(deliveryEstimate);
                        deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
                        deliveryEstimate.setDurationMinValue(row.getInt("service_duration_min_value") + i);//since i is 0
                        deliveryEstimate.setDurationMaxValue(row.getInt("service_duration_max_value") + i);
                        deliveryEstimate.setDurationMinValueWithLeadTime(row.getInt("service_duration_min_value") + i + sumLeadTimes);
                        deliveryEstimate.setDurationMaxValueWithLeadTime(row.getInt("service_duration_max_value") + i + sumLeadTimes);
                        deliveryCutOff.setDeliveryEstimate(deliveryEstimate);
                        break;
                    }
                } else {
                    com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
                    deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff((availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7, deliveryCutOffTime / 1000000L);
                    if (deliveryCutOffList == null) {
                        deliveryCutOffList = new ArrayList<>();
                    }
                    deliveryCutOffList.add(deliveryCutOff);
                    DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
                    rawDelivery.setDeliveryEstimate(deliveryEstimate);
                    deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
                    deliveryEstimate.setDurationMinValue(row.getInt("service_duration_min_value") + i);
                    deliveryEstimate.setDurationMaxValue(row.getInt("service_duration_max_value") + i);
                    deliveryEstimate.setDurationMinValueWithLeadTime(row.getInt("service_duration_min_value") + i + sumLeadTimes);
                    deliveryEstimate.setDurationMaxValueWithLeadTime(row.getInt("service_duration_max_value") + i + sumLeadTimes);
                    deliveryCutOff.setDeliveryEstimate(deliveryEstimate);
                    break;
                }
            }
        }
        if (deliveryCutOffList != null)
            rawDelivery.setDeliveryCutOff(deliveryCutOffList);

        //now that we are done creating the new RawDelivery Object
        //check which rawDelivery is quicker. if the new rawDelivery has valid values we are good
        if (rawDelivery.getDeliveryCutOff() != null && rawDelivery.getDeliveryCutOff().size() > 0 && rawDelivery.getDeliveryEstimate() != null) {
            //first check which has lower durations
            String existingDesignator = existingRawDelivery.getDeliveryEstimate().getDurationDesignator();
            String currentDesignator = existingRawDelivery.getDeliveryEstimate().getDurationDesignator();
            long existCommonDenom = 0;long currentCommonDenom = 0;
            if (existingDesignator.compareToIgnoreCase("Days")==0) {//convert to minutes
                //existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValue() * 1440;
                existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() * 1440;
            }
            else if (existingDesignator.compareToIgnoreCase("Hours")==0)// hours to inutes
                existCommonDenom = existingRawDelivery.getDeliveryEstimate().getDurationMinValue() * 60;
            else
                existCommonDenom =  existingRawDelivery.getDeliveryEstimate().getDurationMinValue();

            if (currentDesignator.compareToIgnoreCase("Days")==0) {//convert to minutes
               // currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValue() * 1440;
                currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() * 1440;
            }
            else if (currentDesignator.compareToIgnoreCase("Hours")==0)// hours to inutes
                currentCommonDenom = rawDelivery.getDeliveryEstimate().getDurationMinValue() * 60;
            else
                currentCommonDenom =  rawDelivery.getDeliveryEstimate().getDurationMinValue();

            if (isDeliveryQuicker(rawDelivery, existingRawDelivery)) { // new rawDelivery is quicker
                //use this new rawDelivery
                existingRawDeliveryList.remove(0);
                com.genpt.nsight.v4.model.RawDelivery newRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                existingRawDeliveryList.add(newRawDelivery);
            } else if (currentCommonDenom == existCommonDenom) { // durations are equal
                // then check which which is quicker
                //if (rawDelivery.getDeliveryCutOff().get(0) != null && existingRawDelivery.getDeliveryCutOff().get(0) != null && rawDelivery.getDeliveryCutOff().get(0).getCutOffTime() > existingRawDelivery.getDeliveryCutOff().get(0).getCutOffTime()) {
                if (rawDelivery.getDeliveryCutOff().get(0) != null &&
                        existingRawDelivery.getDeliveryCutOff().get(0) != null &&
                        availabilityServiceContextV3.cutOffTimeAfterMax(rawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), existingRawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), shipFromTZ, shipFromTZ)
                        ) {
                    /// use the new one
                    existingRawDeliveryList.remove(0);
                    com.genpt.nsight.v4.model.RawDelivery newRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                    existingRawDeliveryList.add(newRawDelivery);
                }
            }
        }

        //at this point we have the quickest rawDelivery across paid and free for this site
        if (existingPaidRawDelivery != null && isPaidDelivery){
            if (rawDelivery.getDeliveryCutOff() != null && rawDelivery.getDeliveryCutOff().size() > 0 && rawDelivery.getDeliveryEstimate() != null) {
                //first check which has lower durations
                if (rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() < existingPaidRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime()) { // new rawDelivery is quicker
                    //use this rawDelivery.  the old one from the list
                    existingPaidRawDeliveryList.remove(0);
                    com.genpt.nsight.v4.model.RawDelivery paidRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                    existingPaidRawDeliveryList.add(paidRawDelivery);
                } else if (rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() == existingPaidRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime()) { // durations are equal
                    // then check which which is quicker
                    if (rawDelivery.getDeliveryCutOff().get(0) != null
                            && existingPaidRawDelivery.getDeliveryCutOff().get(0) != null
                            && availabilityServiceContextV3.cutOffTimeAfterMax(rawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), existingPaidRawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), shipFromTZ, shipFromTZ)
                            ) {
                        /// use the new one
                        existingPaidRawDeliveryList.remove(0);
                        com.genpt.nsight.v4.model.RawDelivery paidRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                        existingPaidRawDeliveryList.add(paidRawDelivery);
                    }
                }
            }
        }// now we have the quickest paid rawDelivery

        if (existingFreeRawDelivery != null && !isPaidDelivery){
            if (rawDelivery.getDeliveryCutOff() != null && rawDelivery.getDeliveryCutOff().size() > 0 && rawDelivery.getDeliveryEstimate() != null) {
                //first check which has lower durations
                if (rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() < existingFreeRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime()) { // new rawDelivery is quicker
                    //use this new rawDelivery.
                    existingFreeRawDeliveryList.remove(0);
                    com.genpt.nsight.v4.model.RawDelivery freeRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);

                    existingFreeRawDeliveryList.add(freeRawDelivery);
                } else if (rawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime() == existingFreeRawDelivery.getDeliveryEstimate().getDurationMinValueWithLeadTime()) { // durations are equal
                    // then check which which is quicker
                    if (rawDelivery.getDeliveryCutOff().get(0) != null
                            && existingFreeRawDelivery.getDeliveryCutOff().get(0) != null
                            && availabilityServiceContextV3.cutOffTimeAfterMax(rawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), existingFreeRawDelivery.getDeliveryCutOff().get(0).getCutOffTime(), shipFromTZ, shipFromTZ)
                            ) {
                        /// use the new one
                        existingFreeRawDeliveryList.remove(0);
                        com.genpt.nsight.v4.model.RawDelivery freeRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                        existingFreeRawDeliveryList.add(freeRawDelivery);
                    }
                }
            }
        }// now we have the quickest free rawDelivery

        if (isPaidDelivery){
            if (existingPaidRawDeliveryList.size() == 0 && rawDelivery.getDeliveryCutOff()!= null && rawDelivery.getDeliveryCutOff().size() > 0 ){
                com.genpt.nsight.v4.model.RawDelivery paidRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                existingPaidRawDeliveryList.add(paidRawDelivery);
            }
        }else{
            if (existingFreeRawDeliveryList.size() == 0 && rawDelivery.getDeliveryCutOff()!= null && rawDelivery.getDeliveryCutOff().size() > 0){
                com.genpt.nsight.v4.model.RawDelivery freeRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                existingFreeRawDeliveryList.add(freeRawDelivery);
            }
        }

    }

    private void createFreePaidDelivery(String serviceDurationDesignator,Row row,  Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> quickestFreePaidDeliveryBySite, AvailabilityServiceContextV4 availabilityServiceContextV3, Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> freeDeliveryBySite,   Map<String, List<com.genpt.nsight.v4.model.RawDelivery>> paidDeliveryBySite ){
        SplunkLogger splunkLogger = new SplunkLogger(availabilityServiceContextV3.getRequest().getHeaderRequest());

        if (serviceDurationDesignator != null) {
            List<com.genpt.nsight.v4.model.RawDelivery> existingFreeRawDeliveryList = null;
            List<com.genpt.nsight.v4.model.RawDelivery> existingPaidRawDeliveryList =null ;
            List<com.genpt.nsight.v4.model.RawDelivery> existingRawDeliveryList = null;

            boolean isDeliveryPaid = false;
            String deliveryServiceType = row.getString("delivery_service_type");
            if (nSightProperties.getPaidMOTS().contains(deliveryServiceType))
                isDeliveryPaid=true;

            String fromSiteID = row.getUUID("ship_from_site_identifier").toString();
            float shipFromTZ = 0;//set the timezone for the shipfrom site
            if (nSightCache.getSiteIdToSiteMap().get(fromSiteID) != null && nSightCache.getSiteIdToSiteMap().get(fromSiteID).getTimeZone() != null && nSightCache.getSiteIdToSiteMap().get(fromSiteID).getTimeZone().trim() != "")
                try {
                	//this should actually be the TZ of the From site store since the cutoff time needs to be translated to ship from store local time
                	shipFromTZ = Float.parseFloat(nSightCache.getSiteIdToSiteMap().get(fromSiteID).getTimeZone());
                }catch (Exception ex){
                    LOGGER.error(splunkLogger.toString() + " , Error in createFreePaidDelivery, Message :  {}, StackTrace: {} ", ex.getMessage(), ex.getStackTrace());
                }

            boolean noCutoffTimes = false;
            if (quickestFreePaidDeliveryBySite.containsKey(fromSiteID)) {// entry for this site exists in map
                //get the existing rawDelivery to compare with the new rawDelivery
                existingRawDeliveryList = quickestFreePaidDeliveryBySite.get(fromSiteID);
                existingFreeRawDeliveryList = freeDeliveryBySite.get(fromSiteID);
                existingPaidRawDeliveryList = paidDeliveryBySite.get(fromSiteID);
                if (existingFreeRawDeliveryList == null)
                    existingFreeRawDeliveryList =  new ArrayList<>();
                if (existingPaidRawDeliveryList == null)
                    existingPaidRawDeliveryList =  new ArrayList<>();

                compareAndSetQuickestDeliveryList(availabilityServiceContextV3,row, existingRawDeliveryList,
                        deliveryServiceType,fromSiteID,serviceDurationDesignator,
                        shipFromTZ, isDeliveryPaid, existingFreeRawDeliveryList, existingPaidRawDeliveryList);
                //at this point free, paid and quickest lists would be populated for this site
            } else {
                existingRawDeliveryList = new ArrayList<>();

                //found the first rawDelivery
                com.genpt.nsight.v4.model.RawDelivery rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
                rawDelivery.setDeliveryType(deliveryServiceType);
                if (row.getColumnDefinitions().contains("delivery_service_zone"))
                    rawDelivery.setZone(row.getString("delivery_service_zone"));
                rawDelivery.setSiteId(fromSiteID);

                List<com.genpt.nsight.v4.model.DeliveryCutOff> deliveryCutOffList = null;

                for (int i = 0; i < 7; i++) {
                    //check if the time is a null value
                    boolean isNull = row.isNull(dayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7]);
                    //include lead times in calculation
                    String dayOfWeek = leadTimesDayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7];
                    int sumLeadTimes = availabilityServiceContextV3.getLeadTimesForSite(dayOfWeek, fromSiteID, deliveryServiceType);

                    if (!isNull) {
                        long deliveryCutOffTime = row.getTime(dayOfWeekColNames[(availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7]);
                        boolean afterCutOff = false;
                        if (availabilityServiceContextV3.getStoreOffSetTime() != null && shipFromTZ != 0 && nSightCache.getSiteIdToSiteMap().get(fromSiteID)!=null && nSightCache.getSiteIdToSiteMap().get(fromSiteID).getTimeZoneID()!=null) {
                        	if (!availabilityServiceContextV3.requestTimeBeforeCutOffTime(deliveryCutOffTime/1000000, shipFromTZ,row.getInt("service_duration_min_value")+sumLeadTimes,nSightCache.getSiteIdToSiteMap().get(fromSiteID)))//expects millis
                                afterCutOff=true;
                        }else{
                            long now = 1000000 * availabilityServiceContextV3.getTimeOfDay();
                            if (deliveryCutOffTime < now)
                                afterCutOff=true;
                        }

                        //if we have passed the cutoff time
                        if (i == 0 ) {
                            if (afterCutOff) {
                                continue;//passed the cutoff time for today
                            }
                            else{//found before the cutoff time for today
                                com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
                                deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff((availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7, deliveryCutOffTime / 1000000L);
                                if (deliveryCutOffList == null) {
                                    deliveryCutOffList = new ArrayList<>();
                                }
                                deliveryCutOffList.add(deliveryCutOff);
                                DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
                                rawDelivery.setDeliveryEstimate(deliveryEstimate);
                                deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
                                deliveryEstimate.setDurationMinValue(row.getInt("service_duration_min_value") + i);
                                deliveryEstimate.setDurationMaxValue(row.getInt("service_duration_max_value") + i);
                                deliveryEstimate.setDurationMinValueWithLeadTime(row.getInt("service_duration_min_value") + i + sumLeadTimes);
                                deliveryEstimate.setDurationMaxValueWithLeadTime(row.getInt("service_duration_max_value") + i + sumLeadTimes);
                                deliveryCutOff.setDeliveryEstimate(deliveryEstimate);
                                break;
                            }
                        } else {
                            com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
                            deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff((availabilityServiceContextV3.getDayOfWeekLocal() + i) % 7, deliveryCutOffTime / 1000000L);
                            if (deliveryCutOffList == null) {
                                deliveryCutOffList = new ArrayList<>();
                            }
                            deliveryCutOffList.add(deliveryCutOff);
                            DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
                            rawDelivery.setDeliveryEstimate(deliveryEstimate);
                            deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
                            deliveryEstimate.setDurationMinValue(row.getInt("service_duration_min_value") + i);
                            deliveryEstimate.setDurationMaxValue(row.getInt("service_duration_max_value") + i);
                            deliveryEstimate.setDurationMinValueWithLeadTime(row.getInt("service_duration_min_value") + i + sumLeadTimes);
                            deliveryEstimate.setDurationMaxValueWithLeadTime(row.getInt("service_duration_max_value") + i + sumLeadTimes);
                            deliveryCutOff.setDeliveryEstimate(deliveryEstimate);
                            break;
                        }
                    }
                }//loop cutoffs

                if (deliveryCutOffList != null && deliveryCutOffList.size() > 0)
                    rawDelivery.setDeliveryCutOff(deliveryCutOffList);

                if (rawDelivery.getDeliveryCutOff()!= null && rawDelivery.getDeliveryCutOff().size() > 0) {
                    existingRawDeliveryList.add(rawDelivery);

                    if (isDeliveryPaid ) {
                        existingPaidRawDeliveryList = new ArrayList<>();
                        com.genpt.nsight.v4.model.RawDelivery paidRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                        existingPaidRawDeliveryList.add(paidRawDelivery);
                    }
                    else {
                        existingFreeRawDeliveryList = new ArrayList<>();
                        com.genpt.nsight.v4.model.RawDelivery freeRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                        existingFreeRawDeliveryList.add(freeRawDelivery);
                    }
                }
                //finished adding the first found rawDelivery
            }//end of else already have rawDelivery list for this site

            /// finally add it to the map, if the list is populated
            if (existingRawDeliveryList !=null && existingRawDeliveryList.size() > 0)
                quickestFreePaidDeliveryBySite.put(fromSiteID, existingRawDeliveryList);

            if (isDeliveryPaid) {
                if (existingPaidRawDeliveryList != null) {
                    paidDeliveryBySite.put(fromSiteID, existingPaidRawDeliveryList);
                }
            }
            else {
                if (existingFreeRawDeliveryList != null) {
                    freeDeliveryBySite.put(fromSiteID, existingFreeRawDeliveryList);
                }
            }

        }//end of check service duration designator
    }

    private com.genpt.nsight.v4.model.RawDelivery createDelivery(String serviceDurationDesignator,Row row, AvailabilityServiceContextV4 availabilityServiceContextV3, List<com.genpt.nsight.v4.model.RawDelivery> rawDeliveryList, Set<String> fromSitesToZip){
        com.genpt.nsight.v4.model.RawDelivery rawDelivery =null;

        if (serviceDurationDesignator != null) {
            rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();

            rawDelivery.setDeliveryType(row.getString("delivery_service_type"));
            if (row.getColumnDefinitions().contains("delivery_service_zone"))
                rawDelivery.setZone(row.getString("delivery_service_zone"));

            String shipFrom = null;
            final String deliveryType = rawDelivery.getDeliveryType();
            rawDelivery.setSiteId(row.getUUID("ship_from_site_identifier").toString());
            shipFrom = row.getUUID("ship_from_site_identifier").toString();
            String shipTo = null;
            if (row.getColumnDefinitions().contains("ship_to_site_identifier"))
                shipTo = row.getUUID("ship_to_site_identifier").toString();
            String shipToZip = null;
            if (row.getColumnDefinitions().contains("ship_to_zip_code"))
                shipToZip = row.getString("ship_to_zip_code");

            //create the delivery estimate
            DeliveryEstimate deliveryEstimate= new DeliveryEstimate();
            deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
            deliveryEstimate.setDurationMinValue(row.getInt("service_duration_min_value"));
            deliveryEstimate.setDurationMaxValue(row.getInt("service_duration_max_value"));
            rawDelivery.setDeliveryEstimate(deliveryEstimate);

            List<com.genpt.nsight.v4.model.DeliveryCutOff> deliveryCutOffList = null;
            for (int i = 0; i < dayOfWeekColNames.length; i++) {
                boolean isNull = row.isNull(dayOfWeekColNames[i]);
                if (!isNull) {
                    long deliveryCutOffTime = row.getTime(dayOfWeekColNames[i]);
                    com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
                    //if (deliveryCutOffTime != 0L)
                    {
                        deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff(i, deliveryCutOffTime / 1000000L);
                        String dayOfWeek = leadTimesDayOfWeekColNames[i];

                        int sumLeadTimes = availabilityServiceContextV3.getLeadTimesForSite(dayOfWeek, shipFrom, deliveryType);
                        deliveryEstimate.setDurationMinValueWithLeadTime(row.getInt("service_duration_min_value") + sumLeadTimes);
                        deliveryEstimate.setDurationMaxValueWithLeadTime(row.getInt("service_duration_max_value")+sumLeadTimes);

                        deliveryCutOff.setDeliveryEstimate(deliveryEstimate);
                        if (deliveryCutOffList == null) {
                            deliveryCutOffList = new ArrayList<>();
                        }
                        deliveryCutOffList.add(deliveryCutOff);
                    }
                }
            }
            if (deliveryCutOffList != null) {
                rawDelivery.setDeliveryCutOff(deliveryCutOffList);
            }

            DeliveryDestination deliveryDestination = new DeliveryDestination();
            if (shipTo != null) {//site. need to get alias and value
                String aliasFullName = availabilityServiceContextV3.getReverseShipToSitesMap().get(shipTo);
                if (aliasFullName != null) {
                    String arr[] = aliasFullName.split("\\|");
                    if (arr.length > 1) {
                        deliveryDestination.setValue(arr[1]);
                        deliveryDestination.setName(arr[0]);
                        rawDelivery.setDeliveryDestination(deliveryDestination);
                        rawDeliveryList.add(rawDelivery);//add it to the list
                    }//valid alias names for SiteID
                }
            }
            if (shipToZip != null && shipToZip.length() == 5) {//zip, that means they had requested a zipcode
                //zipcode  5 digit
                deliveryDestination.setName("Zip Code");
                deliveryDestination.setValue(shipToZip);
                rawDelivery.setDeliveryDestination(deliveryDestination);
                rawDeliveryList.add(rawDelivery);//add it to the list
                //also add it to the set which stores 5 digit deliveris
                fromSitesToZip.add(rawDelivery.getDeliveryType()+shipFrom+shipToZip.substring(0,3));
                //System.out.println("Added zip: " + rawDelivery.getDeliveryType()+shipFrom+shipToZip.substring(0,3));
            }

            if (shipToZip != null && shipToZip.length() == 3) {//3 digit zipcode
                //check if the 5 digit for this 3 digit is already present
                //if its present ignore the 3 digit
                if (fromSitesToZip.contains(rawDelivery.getDeliveryType() + shipFrom + shipToZip)) {
                    //ignore this zipcode
                    //System.out.println("Ignoring ZIP: "+ rawDelivery.getDeliveryType() + shipFrom + shipToZip);
                } else {
                    //loop through the requested 5 digit zips
                    for (String zipcode : availabilityServiceContextV3.getZipCodesSet()) {
                        if (zipcode.contains(shipToZip)) {
                            //create a new deliuvery section for this zip
                            com.genpt.nsight.v4.model.RawDelivery altRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                            DeliveryDestination deliveryAltDestination = new DeliveryDestination();
                            deliveryAltDestination.setName("Zip Code");
                            deliveryAltDestination.setValue(zipcode);
                            altRawDelivery.setDeliveryDestination(deliveryAltDestination);
                            if (!rawDeliveryList.contains(altRawDelivery))
                                rawDeliveryList.add(altRawDelivery);//add it to the list
                        }
                    }
                }
            }
        }

        return rawDelivery;
    }

    //added for V4 implemntation. creates a free delivery and adds it to an existing list
    private com.genpt.nsight.v4.model.RawDelivery createFreeDelivery(String serviceDurationDesignator,Row row, AvailabilityServiceContextV4 availabilityServiceContextV3, List<com.genpt.nsight.v4.model.RawDelivery> rawDeliveryList, Set<String> fromSitesToZip){
        com.genpt.nsight.v4.model.RawDelivery rawDelivery =null;

        if (serviceDurationDesignator != null) {
            int serviceDurationMin = row.getInt("service_duration_min_value");
            int serviceDurationMax = row.getInt("service_duration_max_value");
            String deliveryCategory = row.getString("delivery_category");
            String deliveryDay = row.getString("day");
            long deliveryCutOffTime = row.getTime("cutoff");
            //Need these 2 new fields for On Demand
            long earliestTime = -1;
            long latestTime = -1;
            if (!row.isNull("earliest_time"))
                earliestTime = row.getTime("earliest_time")/ 1000000L;
            if (!row.isNull("latest_time"))
                latestTime =row.getTime("latest_time")/ 1000000L;

            com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
            List<com.genpt.nsight.v4.model.DeliveryCutOff> deliveryCutOffList = null;

            String shipFrom = null;
           final String deliveryType = row.getString("delivery_service_type");
            shipFrom = row.getUUID("ship_from_site_identifier").toString();
            String shipTo = null;
            if (row.getColumnDefinitions().contains("ship_to_site_identifier"))
                shipTo = row.getUUID("ship_to_site_identifier").toString();
            String shipToZip = null;
            if (row.getColumnDefinitions().contains("ship_to_zip_code"))
                shipToZip = row.getString("ship_to_zip_code");
            deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff(Arrays.asList(DeliveryCutOffSerializer.DAY_OF_WEEK).indexOf(deliveryDay), deliveryCutOffTime / 1000000L);
            //we have two new fields earliest time and latest time for On Demand
            if (earliestTime != -1)
                deliveryCutOff.setEarliesTime(earliestTime);
            if (latestTime != -1)
                deliveryCutOff.setLatestTime(latestTime);
            List<com.genpt.nsight.v4.model.RawDelivery> existRawDeliveryList = rawDeliveryList;
            boolean addToList = true;

            int sumLeadTimes = availabilityServiceContextV3.getLeadTimesForSite(deliveryDay, shipFrom, deliveryType);

            DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
            //if its not days then lead times need to be converted to the same since lead times are in days
            if (sumLeadTimes > 0  && (serviceDurationDesignator.compareToIgnoreCase("Minutes")==0 || serviceDurationDesignator.compareToIgnoreCase("Hours")==0) ){
                if ( serviceDurationDesignator.compareToIgnoreCase("Hours")==0)
                    sumLeadTimes = sumLeadTimes*24;
                if ( serviceDurationDesignator.compareToIgnoreCase("Minutes")==0)
                    sumLeadTimes = sumLeadTimes*24*60;
            }

            deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
            deliveryEstimate.setDurationMinValue(serviceDurationMin);
            deliveryEstimate.setDurationMaxValue(serviceDurationMax);
            deliveryEstimate.setDurationMinValueWithLeadTime(serviceDurationMin+ sumLeadTimes);
            deliveryEstimate.setDurationMaxValueWithLeadTime(serviceDurationMax+sumLeadTimes);
            //rawDelivery.setDeliveryEstimate(deliveryEstimate);
            deliveryCutOff.setDeliveryEstimate(deliveryEstimate);

            if (existRawDeliveryList != null){
                //add it for that site if it exists for that type
/*                for (com.genpt.nsight.v4.model.RawDelivery existRawDelivery : existRawDeliveryList){
                    if (existRawDelivery.getDeliveryType().compareToIgnoreCase(deliveryType)==0){
                        deliveryCutOffList = existRawDelivery.getDeliveryCutOff();
                        deliveryCutOffList.add(deliveryCutOff);
                        rawDelivery = existRawDelivery;
                        rawDelivery.setDeliveryCutOff(deliveryCutOffList);
                        addToList = false;
                        break;
                    }
                }*/
                if (addToList){// this rawDelivery type wasnt in the list. create a new one
                    rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
                    rawDelivery.setDeliveryType(row.getString("delivery_service_type"));
                    rawDelivery.setSiteId(row.getUUID("ship_from_site_identifier").toString());
                    rawDelivery.setDeliveryCategory(deliveryCategory);
                    rawDelivery.setDeliveryEstimate(deliveryEstimate);
                    deliveryCutOffList = new ArrayList<>();
                    deliveryCutOffList.add(deliveryCutOff);
                    rawDelivery.setDeliveryCutOff(deliveryCutOffList);
                }
            }else{// add this new rawDelivery to the list. need to create the list too
                //existRawDeliveryList = new ArrayList<>();
                rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
                rawDelivery.setDeliveryType(row.getString("delivery_service_type"));
                rawDelivery.setSiteId(row.getUUID("ship_from_site_identifier").toString());
                rawDelivery.setDeliveryCategory(deliveryCategory);
                rawDelivery.setDeliveryEstimate(deliveryEstimate);
                deliveryCutOffList = new ArrayList<>();
                deliveryCutOffList.add(deliveryCutOff);
                rawDelivery.setDeliveryCutOff(deliveryCutOffList);
            }

            DeliveryDestination deliveryDestination = new DeliveryDestination();
            if (shipTo != null) {//site. need to get alias and value
                String aliasFullName = availabilityServiceContextV3.getReverseShipToSitesMap().get(shipTo);
                if (aliasFullName != null) {
                    String arr[] = aliasFullName.split("\\|");
                    if (arr.length > 1) {
                        deliveryDestination.setValue(arr[1]);
                        deliveryDestination.setName(arr[0]);
                        rawDelivery.setDeliveryDestination(deliveryDestination);
                        if (addToList)
                            rawDeliveryList.add(rawDelivery);//add it to the list
                    }//valid alias names for SiteID
                }
            }
            if (shipToZip != null && shipToZip.length() == 5) {//zip, that means they had requested a zipcode
                //zipcode  5 digit
                deliveryDestination.setName("Zip Code");
                deliveryDestination.setValue(shipToZip);
                rawDelivery.setDeliveryDestination(deliveryDestination);
                rawDeliveryList.add(rawDelivery);//add it to the list
                //also add it to the set which stores 5 digit deliveris
                fromSitesToZip.add(rawDelivery.getDeliveryType()+shipFrom+shipToZip.substring(0,3));
                //System.out.println("Added zip: " + rawDelivery.getDeliveryType()+shipFrom+shipToZip.substring(0,3));
            }

            if (shipToZip != null && shipToZip.length() == 3) {//3 digit zipcode
                //check if the 5 digit for this 3 digit is already present
                //if its present ignore the 3 digit
                if (fromSitesToZip.contains(rawDelivery.getDeliveryType() + shipFrom + shipToZip)) {
                    //ignore this zipcode
                    //System.out.println("Ignoring ZIP: "+ rawDelivery.getDeliveryType() + shipFrom + shipToZip);
                } else {
                    //loop through the requested 5 digit zips
                    for (String zipcode : availabilityServiceContextV3.getZipCodesSet()) {
                        if (zipcode.contains(shipToZip)) {
                            //create a new deliuvery section for this zip
                            com.genpt.nsight.v4.model.RawDelivery altRawDelivery = (com.genpt.nsight.v4.model.RawDelivery) JavaDeepClone.deepClone(rawDelivery);
                            DeliveryDestination deliveryAltDestination = new DeliveryDestination();
                            deliveryAltDestination.setName("Zip Code");
                            deliveryAltDestination.setValue(zipcode);
                            altRawDelivery.setDeliveryDestination(deliveryAltDestination);
                            if (!rawDeliveryList.contains(altRawDelivery))
                                rawDeliveryList.add(altRawDelivery);//add it to the list
                        }
                    }
                }
            }
        }

        return rawDelivery;
    }

    private com.genpt.nsight.v4.model.RawDelivery createFreeDeliveryGCP(String serviceDurationDesignator, DeliveryOption deliveryOption, AvailabilityServiceContextV4 availabilityServiceContextV3, List<com.genpt.nsight.v4.model.RawDelivery> rawDeliveryList){
        com.genpt.nsight.v4.model.RawDelivery rawDelivery =null;

        if (serviceDurationDesignator != null) {
            int serviceDurationMin = deliveryOption.getDurationMinValue();
            int serviceDurationMax = deliveryOption.getDurationMaxValue();
            String deliveryCategory = deliveryOption.getCategoryType();
            if(deliveryCategory!=null)
                deliveryCategory = WordUtils.capitalize(deliveryCategory.toLowerCase());
            String deliveryDay = deliveryOption.getDay();
            long deliveryCutOffTime=-1;
            if(deliveryOption.getCutoff()!=null){
                String[] arr = deliveryOption.getCutoff().split(":");
                deliveryCutOffTime = (Integer.valueOf(arr[0])*60+Integer.valueOf(arr[1]))*60*1000;
            }

            //Need these 2 new fields for On Demand
            long earliestTime = -1;
            long latestTime = -1;
            if (deliveryOption.getEarliestTime()!=null) {
                String[] arr = deliveryOption.getEarliestTime().split(":");
                earliestTime = (Integer.valueOf(arr[0])*60+Integer.valueOf(arr[1]))*60*1000;
                earliestTime = earliestTime;
            }
            if (deliveryOption.getLatestTime()!=null) {
                String[] arr = deliveryOption.getLatestTime().split(":");
                latestTime = (Integer.valueOf(arr[0])*60+Integer.valueOf(arr[1]))*60*1000;
                latestTime = latestTime;
            }
            com.genpt.nsight.v4.model.DeliveryCutOff deliveryCutOff = null;
            List<com.genpt.nsight.v4.model.DeliveryCutOff> deliveryCutOffList = null;
            final String deliveryType = deliveryOption.getServiceType();
            String shipFrom = null;
            shipFrom = nSightCache.getAliasToSiteIdMap().get("9 Digit Store Number|"+deliveryOption.getSource());
            if(shipFrom==null)
                shipFrom = nSightCache.getAliasToSiteIdMap().get("DC Abbreviation Code|"+deliveryOption.getSource());
            if(shipFrom==null)
                shipFrom = nSightCache.getAliasToSiteIdMap().get("Supplier Abbreviation Location|"+deliveryOption.getSource());
            String shipTo = null;
            shipTo = nSightCache.getAliasToSiteIdMap().get("9 Digit Store Number|"+deliveryOption.getDestination());
            if(shipTo==null)
                shipTo = nSightCache.getAliasToSiteIdMap().get("DC Abbreviation Code|"+deliveryOption.getDestination());
            if(shipTo==null)
                shipTo = nSightCache.getAliasToSiteIdMap().get("Supplier Abbreviation Location|"+deliveryOption.getDestination());

            deliveryCutOff = new com.genpt.nsight.v4.model.DeliveryCutOff(Arrays.asList(DeliveryCutOffSerializer.DAY_OF_WEEK_UPPER).indexOf(deliveryDay.toUpperCase()), deliveryCutOffTime);
            //we have two new fields earliest time and latest time for On Demand
            if (earliestTime != -1)
                deliveryCutOff.setEarliesTime(earliestTime);
            if (latestTime != -1)
                deliveryCutOff.setLatestTime(latestTime);
            List<com.genpt.nsight.v4.model.RawDelivery> existRawDeliveryList = rawDeliveryList;
            boolean addToList = true;

            int sumLeadTimes = availabilityServiceContextV3.getLeadTimesForSite(deliveryDay, shipFrom, deliveryType);

            DeliveryEstimate deliveryEstimate = new DeliveryEstimate();
            //if its not days then lead times need to be converted to the same since lead times are in days
            if (sumLeadTimes > 0  && (serviceDurationDesignator.compareToIgnoreCase("Minutes")==0 || serviceDurationDesignator.compareToIgnoreCase("Hours")==0) ){
                if ( serviceDurationDesignator.compareToIgnoreCase("Hours")==0)
                    sumLeadTimes = sumLeadTimes*24;
                if ( serviceDurationDesignator.compareToIgnoreCase("Minutes")==0)
                    sumLeadTimes = sumLeadTimes*24*60;
            }

            deliveryEstimate.setDurationDesignator(serviceDurationDesignator);
            deliveryEstimate.setDurationMinValue(serviceDurationMin);
            deliveryEstimate.setDurationMaxValue(serviceDurationMax);
            deliveryEstimate.setDurationMinValueWithLeadTime(serviceDurationMin+ sumLeadTimes);
            deliveryEstimate.setDurationMaxValueWithLeadTime(serviceDurationMax+sumLeadTimes);
            deliveryCutOff.setDeliveryEstimate(deliveryEstimate);

            if (existRawDeliveryList != null){
                //add it for that site if it exists for that type
                if (addToList){// this rawDelivery type wasnt in the list. create a new one
                    rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
                    rawDelivery.setDeliveryType(deliveryOption.getServiceType());
                    rawDelivery.setSiteId(shipFrom);
                    rawDelivery.setDeliveryCategory(deliveryCategory);
                    rawDelivery.setDeliveryEstimate(deliveryEstimate);
                    deliveryCutOffList = new ArrayList<>();
                    deliveryCutOffList.add(deliveryCutOff);
                    rawDelivery.setDeliveryCutOff(deliveryCutOffList);
                }
            }else{// add this new rawDelivery to the list. need to create the list too
                //existRawDeliveryList = new ArrayList<>();
                rawDelivery = new com.genpt.nsight.v4.model.RawDelivery();
                rawDelivery.setDeliveryType(deliveryOption.getServiceType());
                rawDelivery.setSiteId(shipFrom);
                rawDelivery.setDeliveryCategory(deliveryCategory);
                rawDelivery.setDeliveryEstimate(deliveryEstimate);
                deliveryCutOffList = new ArrayList<>();
                deliveryCutOffList.add(deliveryCutOff);
                rawDelivery.setDeliveryCutOff(deliveryCutOffList);
            }

            DeliveryDestination deliveryDestination = new DeliveryDestination();
            if (shipTo != null && shipFrom!=null) {//site. need to get alias and value
                String aliasFullName = availabilityServiceContextV3.getReverseShipToSitesMap().get(shipTo);
                if (aliasFullName != null) {
                    String arr[] = aliasFullName.split("\\|");
                    if (arr.length > 1) {
                        deliveryDestination.setValue(arr[1]);
                        deliveryDestination.setName(arr[0]);
                        rawDelivery.setDeliveryDestination(deliveryDestination);
                        if (addToList)
                            rawDeliveryList.add(rawDelivery);//add it to the list
                    }//valid alias names for SiteID
                }
            }
        }

        return rawDelivery;
    }

}
